GCCPDU = """

-- ASN module extracted from ITU-T T.124 (01/2007)

-- Module GCC-PROTOCOL (T.127:08/2007)
GCC-PROTOCOL {itu-t(0) recommendation(0) t(20) t124(124) version(0) 2 asn1Modules(2) gcc-protocol(1)} 
DEFINITIONS AUTOMATIC TAGS ::=
BEGIN

-- Export all symbols
-- =============================================================================
-- Part 1:  Elements of which messages are composed
-- =============================================================================
ChannelID ::= INTEGER(1..65535)

StaticChannelID ::= INTEGER(1..1000)

-- Those assigned by specifications
DynamicChannelID ::= INTEGER(1001..65535)

-- Those created and deleted by MCS
UserID ::= DynamicChannelID

TokenID ::= INTEGER(1..65535)

StaticTokenID ::= INTEGER(1..16383)

-- Those assigned by specifications
DynamicTokenID ::= INTEGER(16384..65535)

-- Those assigned by the registry
Time ::= INTEGER(-2147483648..2147483647)

-- Time in seconds
Handle ::= INTEGER(0..4294967295)

-- 32-bit value
H221NonStandardIdentifier ::= OCTET STRING(SIZE (4..255))

-- First four octets shall be country code and
-- Manufacturer code, assigned as specified in
-- Annex A/H.221 for NS-cap and NS-comm
Key ::= CHOICE -- Identifier of a standard or non-standard object
         {
  object           OBJECT IDENTIFIER,
  h221NonStandard  H221NonStandardIdentifier
}

NonStandardParameter ::= SEQUENCE {key   Key,
                                   data  OCTET STRING
}

TextString ::= BMPString(SIZE (0..255))

-- Basic Multilingual Plane of ISO/IEC 10646-1 (Unicode)
simpleTextFirstCharacter UniversalString ::= {0, 0, 0, 0}

simpleTextLastCharacter UniversalString ::= {0, 0, 0, 255}

SimpleTextString ::=
  BMPString(SIZE (0..255))
    (FROM ("0..255"))

SimpleNumericString ::= NumericString(SIZE (1..255))(FROM ("0123456789"))

DiallingString ::= NumericString(SIZE (1..16))(FROM ("0123456789"))

SubAddressString ::= NumericString(SIZE (1..40))(FROM ("0123456789"))

ExtraDiallingString ::= TextString(SIZE (1..255))(FROM ("0123456789#*,"))

UserData ::= SET OF SEQUENCE {key    Key,
                              value  OCTET STRING OPTIONAL}

Password ::= SEQUENCE {
  numeric         SimpleNumericString,
  text            SimpleTextString OPTIONAL,
  ...,
  unicodeText     TextString OPTIONAL
}

PasswordSelector ::= CHOICE {
  numeric      SimpleNumericString,
  text         SimpleTextString,
  ...,
  unicodeText  TextString
}

ChallengeResponseItem ::= CHOICE {
  passwordString  PasswordSelector,
  responseData    UserData,
  ...
}

ChallengeResponseAlgorithm ::= CHOICE {
  passwordInTheClear    NULL,
  nonStandardAlgorithm  NonStandardParameter,
  ...
}

ChallengeItem ::= SEQUENCE {
  responseAlgorithm  ChallengeResponseAlgorithm,
  challengeData      UserData,
  ...
}

ChallengeRequest ::= SEQUENCE {
  challengeTag  INTEGER,
  challengeSet  SET OF ChallengeItem,
  -- Set of algorithms offered for response
  ...
}

ChallengeResponse ::= SEQUENCE {
  challengeTag       INTEGER,
  responseAlgorithm  ChallengeResponseAlgorithm,
  -- Specific algorithm selected from the set of
  -- items presented in the ChallengeRequest
  responseItem       ChallengeResponseItem,
  ...
}

PasswordChallengeRequestResponse ::= CHOICE {
  passwordInTheClear        PasswordSelector,
  challengeRequestResponse
    SEQUENCE {challengeRequest   ChallengeRequest OPTIONAL,
              challengeResponse  ChallengeResponse OPTIONAL,
              ...},
  ...
}

ConferenceName ::= SEQUENCE {
  numeric         SimpleNumericString,
  text            SimpleTextString OPTIONAL,
  ...,
  unicodeText     TextString OPTIONAL
}

ConferenceNameSelector ::= CHOICE {
  numeric      SimpleNumericString,
  text         SimpleTextString,
  ...,
  unicodeText  TextString
}

ConferenceNameModifier ::= SimpleNumericString

Privilege ::= ENUMERATED {
  terminate(0), ejectUser(1), add(2), lockUnlock(3), transfer(4), ...
  }

TerminationMethod ::= ENUMERATED {automatic(0), manual(1), ...
                                  }

ConferencePriorityScheme ::= CHOICE {
  nonStandardScheme  NonStandardParameter,
  ...
}

ConferencePriority ::= SEQUENCE {
  priority  INTEGER(0..65535),
  scheme    ConferencePriorityScheme,
  ...
}

NodeCategory ::= CHOICE {
  conventional         NULL,
  counted              NULL,
  anonymous            NULL,
  nonStandardCategory  NonStandardParameter,
  ...
}

ConferenceMode ::= CHOICE {
  conventional-only     NULL,
  counted-only          NULL,
  anonymous-only        NULL,
  conventional-control  NULL,
  unrestricted-mode     NULL,
  non-standard-mode     NonStandardParameter,
  ...
}

NetworkAddress ::=
  SEQUENCE (SIZE (1..64)) OF
    CHOICE -- Listed in order of use
     {aggregatedChannel
        SEQUENCE {transferModes
                    SEQUENCE -- One or more-- {speech         BOOLEAN,
                                              voice-band     BOOLEAN,
                                              digital-56k    BOOLEAN,
                                              digital-64k    BOOLEAN,
                                              digital-128k   BOOLEAN,
                                              digital-192k   BOOLEAN,
                                              digital-256k   BOOLEAN,
                                              digital-320k   BOOLEAN,
                                              digital-384k   BOOLEAN,
                                              digital-512k   BOOLEAN,
                                              digital-768k   BOOLEAN,
                                              digital-1152k  BOOLEAN,
                                              digital-1472k  BOOLEAN,
                                              digital-1536k  BOOLEAN,
                                              digital-1920k  BOOLEAN,
                                              packet-mode    BOOLEAN,
                                              frame-mode     BOOLEAN,
                                              atm            BOOLEAN,
                                              ...},
                  internationalNumber     DiallingString,
                  subAddress              SubAddressString OPTIONAL,
                  extraDialling           ExtraDiallingString OPTIONAL,
                  highLayerCompatibility
                    SEQUENCE {telephony3kHz    BOOLEAN,
                              telephony7kHz    BOOLEAN,
                              videotelephony   BOOLEAN,
                              videoconference  BOOLEAN,
                              audiographic     BOOLEAN,
                              audiovisual      BOOLEAN,
                              multimedia       BOOLEAN,
                              ...} OPTIONAL,
                  ...},
      transportConnection
        SEQUENCE {nsapAddress        OCTET STRING(SIZE (1..20)),
                  transportSelector  OCTET STRING OPTIONAL},
      nonStandard          NonStandardParameter,
      ...}

MediaList ::= SEQUENCE {audio  BOOLEAN,
                        video  BOOLEAN,
                        data   BOOLEAN,
                        ...
}

ChannelAggregationMethod ::= CHOICE {
  h221           NULL,
  h244           NULL,
  iso-iec-13871  NULL,
  -- The actual mode of bonding is dynamically selected according
  -- to the procedures described in ISO/IEC 13871.
  nonStandard    NonStandardParameter,
  ...
}

Profile ::= CHOICE {
  simpleProfile
    CHOICE {-- Basic transfer modes:
            speech             NULL, -- Simple telephony--
            telephony-3kHz     NULL, -- Rec. G.711--
            telephony-7kHz     NULL, -- Rec. G.722--
            voice-band         NULL, -- Modems--
            frameRelay         NULL,
            -- T.120-only data profiles (Rec. T.123):
            t123-pstn-basic    NULL,
            t123-psdn-basic    NULL,
            t123-b-isdn-basic  NULL},
  multimediaProfile
    SEQUENCE {profile
                CHOICE {h310   NULL,
                        h320   NULL,
                        h321   NULL,
                        h322   NULL,
                        h323   NULL,
                        h324   NULL,
                        h324m  NULL,
                        asvd   NULL,
                        dsvd   NULL},
              t120Data  BOOLEAN},
  dsmccDownloadProfile  NULL,
  nonStandard           NonStandardParameter,
  ...
}

ExtendedE164NetworkAddress ::= SEQUENCE {
  internationalNumber  DiallingString,
  subAddress           SubAddressString OPTIONAL,
  extraDialling        ExtraDiallingString OPTIONAL,
  ...
}

TransportAddress ::= SEQUENCE {
  nsapAddress        OCTET STRING(SIZE (1..20)),
  transportSelector  OCTET STRING OPTIONAL
}

GSTNConnection ::= SEQUENCE {networkAddress  ExtendedE164NetworkAddress,
                             ...
}

ISDNConnection ::= SEQUENCE {
  circuitTypes
    SET OF
      CHOICE {digital-64k         NULL,
              digital-2x64k       NULL,
              digital-384k        NULL,
              digital-1536        NULL,
              digital-1920k       NULL,
              multirate-base-64k  INTEGER(1..30) -- See Note 1
    },
  networkAddress          ExtendedE164NetworkAddress,
  highLayerCompatibility
    SEQUENCE {-- Those are supported code points for IE HLC of the D
              -- protocol (Rec. Q.931).
              telephony3kHz    BOOLEAN,
              telephony7kHz    BOOLEAN,
              videotelephony   BOOLEAN,
              videoconference  BOOLEAN,
              audiographic     BOOLEAN,
              audiovisual      BOOLEAN,
              multimedia       BOOLEAN,
              ...} OPTIONAL,
  ...
}

-- Note 1:	digital-2x64k differs from multirate-base-64k 
-- 			with a multiplier value of 2; 
--			in the first case 
--			the network is requested an 8 kHz integrity with Restricted
--			Differential Time Delay (RDTD); 
--			in the second case
--			the network is requested a Time Slot 
--			Sequence integrity (see 4.5.5/Q.931)

CSDNConnection ::= SEQUENCE {
  circuitTypes    SET OF CHOICE {digital-56k  NULL,
                                 digital-64k  NULL},
  networkAddress  ExtendedE164NetworkAddress,
  ...
}

PSDNConnection ::= SEQUENCE {
  networkAddress
    CHOICE {extendedE164NetworkAddress  ExtendedE164NetworkAddress,
            transportAddress            TransportAddress,
            nonStandard                 NonStandardParameter},
  ...
}

ATMConnection ::= SEQUENCE {
  networkAddress
    CHOICE {extendedE164  ExtendedE164NetworkAddress,
            nsapAddress   TransportAddress,
            -- this case is reserved for NSAPs only: the 
            -- optional transport selector shall never be used
            nonStandard   NonStandardParameter},
  maxTransferRate  INTEGER(0..MAX) OPTIONAL,
  -- in cells per seconds
  ...
}

NetworkConnection ::= CHOICE {
  gstnConnection              GSTNConnection,
  isdnConnection              ISDNConnection,
  csdnConnection              CSDNConnection,
  psdnConnection              PSDNConnection,
  atmConnection               ATMConnection,
  extendedE164NetworkAddress  ExtendedE164NetworkAddress,
  -- Note: LAN connections and leased
  transportAddress            TransportAddress,
  -- lines (Rec. G.703/G.704) may be 
  nonStandard                 NonStandardParameter,
  -- covered by one of these
  ...
}

NetworkAddressV2 ::=
  SET OF
    SEQUENCE {networkConnection
                CHOICE {singleConnection       NetworkConnection,
                        aggregatedConnections
                          SEQUENCE {connectionList
                                      SET (SIZE (1..30)) OF
                                        CHOICE {isdnConnection  ISDNConnection,
                                                csdnConnection  CSDNConnection,
                                                ...},
                                    aggregationMethods
                                      SET OF ChannelAggregationMethod OPTIONAL,
                                    ...}},
              profiles           SET OF Profile OPTIONAL,
              mediaConcerned     MediaList OPTIONAL,
              ...}

NodeType ::= ENUMERATED {terminal(0), multiportTerminal(1), mcu(2), ...
                         }

NodeProperties ::= SEQUENCE {
  managementDevice  BOOLEAN,
  -- Is the node a device such as a reservation system
  peripheralDevice  BOOLEAN,
  -- Is the node a peripheral to a primary node
  ...
}

AsymmetryIndicator ::= CHOICE {
  callingNode  NULL,
  calledNode   NULL,
  unknown      INTEGER(0..4294967295)
  -- Uniformly distributed 32-bit random number
}

AlternativeNodeID ::= CHOICE {h243NodeID  OCTET STRING(SIZE (2)),
                              ...
}

ConferenceDescriptor ::= SEQUENCE {
  conferenceName              ConferenceName,
  conferenceNameModifier      ConferenceNameModifier OPTIONAL,
  conferenceDescription       TextString OPTIONAL,
  lockedConference            BOOLEAN,
  passwordInTheClearRequired  BOOLEAN,
  networkAddress              NetworkAddress OPTIONAL,
  ...,
  defaultConferenceFlag       BOOLEAN,
  conferenceMode              ConferenceMode
}

NodeRecord ::= SEQUENCE {
  superiorNode         UserID OPTIONAL,
  -- Not present only for the Top GCC Provider
  nodeType             NodeType,
  nodeProperties       NodeProperties,
  nodeName             TextString OPTIONAL,
  participantsList     SEQUENCE OF TextString OPTIONAL,
  siteInformation      TextString OPTIONAL,
  networkAddress       NetworkAddress OPTIONAL,
  alternativeNodeID    AlternativeNodeID OPTIONAL,
  userData             UserData OPTIONAL,
  ...,
  nodeCategory         NodeCategory OPTIONAL,
  networkAddressV2     NetworkAddressV2 OPTIONAL
}

SessionKey ::= SEQUENCE {
  applicationProtocolKey  Key,
  sessionID               ChannelID OPTIONAL
}

ChannelType ::= ENUMERATED {
  static(0), dynamicMulticast(1), dynamicPrivate(2), dynamicUserId(3)
}

ApplicationRecord ::= SEQUENCE {
  applicationActive           BOOLEAN,
  -- Active/Inactive flag
  conductingOperationCapable  BOOLEAN,
  -- Maximum one per node per session
  startupChannel              ChannelType OPTIONAL,
  applicationUserID           UserID OPTIONAL,
  -- User ID assigned to the Application Protocol Entity
  nonCollapsingCapabilities
    SET OF
      SEQUENCE {capabilityID     CapabilityID,
                applicationData  OCTET STRING OPTIONAL} OPTIONAL,
  ...
}

CapabilityID ::= CHOICE {
  standard     INTEGER(0..65535),
  -- Assigned by Application Protocol specifications
  nonStandard  Key
}

CapabilityClass ::= CHOICE {
  logical      NULL,
  unsignedMin  INTEGER(0..MAX), -- Capability value
  unsignedMax  INTEGER(0..MAX), -- Capability value
  ...
}

EntityID ::= INTEGER(0..65535)

ApplicationInvokeSpecifier ::= SEQUENCE {
  sessionKey             SessionKey,
  expectedCapabilitySet
    SET OF
      SEQUENCE {capabilityID     CapabilityID,
                capabilityClass  CapabilityClass,
                ...} OPTIONAL,
  startupChannel         ChannelType OPTIONAL,
  mandatoryFlag          BOOLEAN,
  -- TRUE indicates required Application Protocol Entity
  ...
}

RegistryKey ::= SEQUENCE {
  sessionKey  SessionKey,
  resourceID  OCTET STRING(SIZE (0..64))
}

RegistryItem ::= CHOICE {
  channelID  DynamicChannelID,
  tokenID    DynamicTokenID,
  parameter  OCTET STRING(SIZE (0..64)),
  vacant     NULL,
  ...
}

RegistryEntryOwner ::= CHOICE {
  owned
    SEQUENCE {nodeID    UserID, -- Node ID of the owning node--
              entityID  EntityID -- Entity ID of the owning
}, 
-- Appliction Protocol Entity
  notOwned  NULL -- There is no current owner
}

RegistryModificationRights ::= ENUMERATED {owner(0), session(1), public(2)}

-- ============================================================================
-- Part 2:  PDU Messages
-- ============================================================================
UserIDIndication ::= SEQUENCE {tag  INTEGER,
                               ...
}

ConferenceCreateRequest ::=
  SEQUENCE { -- MCS-Connect-Provider request user data
  conferenceName          ConferenceName,
  convenerPassword        Password OPTIONAL,
  password                Password OPTIONAL,
  lockedConference        BOOLEAN,
  listedConference        BOOLEAN,
  conductibleConference   BOOLEAN,
  terminationMethod       TerminationMethod,
  conductorPrivileges     SET OF Privilege OPTIONAL,
  conductedPrivileges     SET OF Privilege OPTIONAL,
  nonConductedPrivileges  SET OF Privilege OPTIONAL,
  conferenceDescription   TextString OPTIONAL,
  callerIdentifier        TextString OPTIONAL,
  userData                UserData OPTIONAL,
  ...,
  conferencePriority      ConferencePriority OPTIONAL,
  conferenceMode          ConferenceMode OPTIONAL
}

ConferenceCreateResponse ::=
  SEQUENCE { -- MCS-Connect-Provider response user data
  nodeID    UserID, -- Node ID of the sending node
  tag       INTEGER,
  result
    ENUMERATED {success(0), userRejected(1), resourcesNotAvailable(2),
                rejectedForSymmetryBreaking(3),
                lockedConferenceNotSupported(4), ...
                },
  userData  UserData OPTIONAL,
  ...
}

ConferenceQueryRequest ::= SEQUENCE { -- MCS-Connect-Provider request user data
  nodeType            NodeType,
  asymmetryIndicator  AsymmetryIndicator OPTIONAL,
  userData            UserData OPTIONAL,
  ...
}

ConferenceQueryResponse ::=
  SEQUENCE { -- MCS-Connect-Provider response user data
  nodeType                     NodeType,
  asymmetryIndicator           AsymmetryIndicator OPTIONAL,
  conferenceList               SET OF ConferenceDescriptor,
  result                       ENUMERATED {success(0), userRejected(1), ...
                                           },
  userData                     UserData OPTIONAL,
  ...,
  waitForInvitationFlag        BOOLEAN OPTIONAL,
  noUnlistedConferenceFlag     BOOLEAN OPTIONAL
}

ConferenceJoinRequest ::=
  SEQUENCE { -- MCS-Connect-Provider request user data as well as
  -- MCS-Send-Data on Node ID Channel of Top GCC sent
  -- by the receiver of the MCS-Connect-Provider
  conferenceName          ConferenceNameSelector OPTIONAL,
  -- Required when part of MCS-Connect-Provider
  conferenceNameModifier  ConferenceNameModifier OPTIONAL,
  tag                     INTEGER OPTIONAL,
  -- Filled in when sent on Node ID Channel of Top GCC
  password                PasswordChallengeRequestResponse OPTIONAL,
  convenerPassword        PasswordSelector OPTIONAL,
  callerIdentifier        TextString OPTIONAL,
  userData                UserData OPTIONAL,
  ...,
  nodeCategory            NodeCategory OPTIONAL
}

ConferenceJoinResponse ::=
  SEQUENCE { -- MCS-Connect-Provider response user data as well as
  -- MCS-Send-Data on Node ID Channel of
  -- the receiver of the MCS-Connect-Provider
  nodeID                      UserID OPTIONAL,
  -- Node ID of directly connected node only
  topNodeID                   UserID,
  -- Node ID of Top GCC Provider
  tag                         INTEGER,
  conferenceNameAlias         ConferenceNameSelector OPTIONAL,
  passwordInTheClearRequired  BOOLEAN,
  lockedConference            BOOLEAN,
  listedConference            BOOLEAN,
  conductibleConference       BOOLEAN,
  terminationMethod           TerminationMethod,
  conductorPrivileges         SET OF Privilege OPTIONAL,
  -- No privilege shall be listed more than once
  conductedPrivileges         SET OF Privilege OPTIONAL,
  -- No privilege shall be listed more than once
  nonConductedPrivileges      SET OF Privilege OPTIONAL,
  -- No privilege shall be listed more than once
  conferenceDescription       TextString OPTIONAL,
  password                    PasswordChallengeRequestResponse OPTIONAL,
  result
    ENUMERATED {success(0), userRejected(1), invalidConference(2),
                invalidPassword(3), invalidConvenerPassword(4),
                challengeResponseRequired(5), invalidChallengeResponse(6), 
                ...
                },
  userData                    UserData OPTIONAL,
  ...,
  nodeCategory                NodeCategory OPTIONAL,
  conferenceMode              ConferenceMode OPTIONAL
}

ConferenceInviteRequest ::=
  SEQUENCE { -- MCS-Connect-Provider request user data
  conferenceName              ConferenceName,
  nodeID                      UserID, -- Node ID of the sending node
  topNodeID                   UserID, -- Node ID of Top GCC Provider
  tag                         INTEGER,
  passwordInTheClearRequired  BOOLEAN,
  lockedConference            BOOLEAN,
  listedConference            BOOLEAN,
  conductibleConference       BOOLEAN,
  terminationMethod           TerminationMethod,
  conductorPrivileges         SET OF Privilege OPTIONAL,
  -- No privilege shall be listed more than once
  conductedPrivileges         SET OF Privilege OPTIONAL,
  -- No privilege shall be listed more than once
  nonConductedPrivileges      SET OF Privilege OPTIONAL,
  -- No privilege shall be listed more than once
  conferenceDescription       TextString OPTIONAL,
  callerIdentifier            TextString OPTIONAL,
  userData                    UserData OPTIONAL,
  ...,
  conferencePriority          ConferencePriority OPTIONAL,
  nodeCategory                NodeCategory OPTIONAL,
  conferenceMode              ConferenceMode OPTIONAL
}

ConferenceInviteResponse ::=
  SEQUENCE { -- MCS-Connect-Provider response user data
  result    ENUMERATED {success(0), userRejected(1), ...
                        },
  userData  UserData OPTIONAL,
  ...
}

ConferenceAddRequest ::=
  SEQUENCE { -- MCS-Send-Data on Node ID Channel of Top GCC or
  -- Node ID Channel of Adding MCU if specified
  networkAddress       NetworkAddress,
  requestingNode       UserID,
  tag                  INTEGER,
  addingMCU            UserID OPTIONAL,
  userData             UserData OPTIONAL,
  ...,
  nodeCategory         NodeCategory OPTIONAL,
  networkAddressV2     NetworkAddressV2
}

ConferenceAddResponse ::=
  SEQUENCE { -- MCS-Send-Data on Node ID Channel of requester
  tag       INTEGER,
  result
    ENUMERATED {success(0), invalidRequester(1), invalidNetworkType(2),
                invalidNetworkAddress(3), addedNodeBusy(4), networkBusy(5),
                noPortsAvailable(6), connectionUnsuccessful(7), ...
                },
  userData  UserData OPTIONAL,
  ...
}

ConferenceLockRequest ::=
  SEQUENCE { -- MCS-Send-Data on Node ID Channel of Top GCC
  -- No parameters
  ...
}

ConferenceLockResponse ::=
  SEQUENCE { -- MCS-Send-Data on Node ID Channel of requester
  result  ENUMERATED {success(0), invalidRequester(1), alreadyLocked(2), ...
                      },
  ...
}

ConferenceLockIndication ::=
  SEQUENCE { -- MCS-Uniform-Send-Data on GCC-Broadcast-Channel
  -- or MCS-Send-Data on Node ID Channel
  -- No parameters
  ...
}

ConferenceUnlockRequest ::=
  SEQUENCE { -- MCS-Send-Data on Node ID Channel of Top GCC
  -- No parameters
  ...
}

ConferenceUnlockResponse ::=
  SEQUENCE { -- MCS-Send-Data on Node ID Channel of requester
  result  ENUMERATED {success(0), invalidRequester(1), alreadyUnlocked(2), ...
                      },
  ...
}

ConferenceUnlockIndication ::=
  SEQUENCE { -- MCS-Uniform-Send-Data on GCC-Broadcast-Channel
  -- or MCS-Send-Data on Node ID Channel
  -- No parameters
  ...
}

ConferenceTerminateRequest ::=
  SEQUENCE { -- MCS-Send-Data on Node ID Channel of Top GCC
  reason  ENUMERATED {userInitiated(0), timedConferenceTermination(1), ...
                      },
  ...
}

ConferenceTerminateResponse ::=
  SEQUENCE { -- MCS-Send-Data on Node ID Channel of requester
  result  ENUMERATED {success(0), invalidRequester(1), ...
                      },
  ...
}

ConferenceTerminateIndication ::=
  SEQUENCE { -- MCS-Uniform-Send-Data on GCC-Broadcast-Channel
  reason  ENUMERATED {userInitiated(0), timedConferenceTermination(1), ...
                      },
  ...
}

ConferenceEjectUserRequest ::=
  SEQUENCE { -- MCS-Send-Data on Node ID Channel of Top GCC
  nodeToEject  UserID, -- Node ID of the node to eject
  reason       ENUMERATED {userInitiated(0), ...
                           },
  ...
}

ConferenceEjectUserResponse ::=
  SEQUENCE { -- MCS-Send-Data on Node ID Channel of requester
  nodeToEject  UserID, -- Node ID of the node to eject
  result
    ENUMERATED {success(0), invalidRequester(1), invalidNode(2), ...
                },
  ...
}

ConferenceEjectUserIndication ::=
  SEQUENCE { -- MCS-Uniform-Send-Data on GCC-Broadcast-Channel
  nodeToEject  UserID, -- Node ID of the node to eject
  reason
    ENUMERATED {userInitiated(0), higherNodeDisconnected(1),
                higherNodeEjected(2), ...
                },
  ...
}

ConferenceTransferRequest ::=
  SEQUENCE { -- MCS-Send-Data on Node ID Channel of Top GCC
  conferenceName          ConferenceNameSelector,
  -- Name of conference to transfer to
  conferenceNameModifier  ConferenceNameModifier OPTIONAL,
  networkAddress          NetworkAddress OPTIONAL,
  transferringNodes       SET (SIZE (1..65536)) OF UserID OPTIONAL,
  password                PasswordSelector OPTIONAL,
  ...,
  networkAddressV2        NetworkAddressV2 OPTIONAL
}

ConferenceTransferResponse ::=
  SEQUENCE { -- MCS-Send-Data on Node ID Channel of requester
  conferenceName          ConferenceNameSelector,
  -- Name of conference to transfer to
  conferenceNameModifier  ConferenceNameModifier OPTIONAL,
  transferringNodes       SET (SIZE (1..65536)) OF UserID OPTIONAL,
  result                  ENUMERATED {success(0), invalidRequester(1), ...
                                      },
  ...
}

ConferenceTransferIndication ::=
  SEQUENCE { -- MCS-Uniform-Send-Data on GCC-Broadcast-Channel
  conferenceName          ConferenceNameSelector,
  -- Name of conference to transfer to
  conferenceNameModifier  ConferenceNameModifier OPTIONAL,
  networkAddress          NetworkAddress OPTIONAL,
  transferringNodes       SET (SIZE (1..65536)) OF UserID OPTIONAL,
  -- List of Node IDs,
  -- not present if destined for all nodes
  password                PasswordSelector OPTIONAL,
  ...,
  networkAddressV2        NetworkAddressV2 OPTIONAL
}

RosterUpdateIndication ::= SEQUENCE { -- MCS-Send-Data on Node ID Channel or
  -- MCS-Uniform-Send-Data on GCC-Broadcast-Channel
  fullRefresh             BOOLEAN,
  -- Conference Roster and all 
  -- ApplicationProtocol Sessions refreshed
  nodeInformation
    SEQUENCE {nodeRecordList
                CHOICE {noChange  NULL,
                        refresh
                          SET (SIZE (1..65536)) OF
                            SEQUENCE
                             -- One for each node in the conference;
                            -- no node shall be listed more than once
                            {nodeID      UserID, -- Node ID of the node--
                             nodeRecord  NodeRecord},
                        update
                          SET (SIZE (1..65536)) OF
                            SEQUENCE
                             -- One for each node changing its node record;
                            -- no node shall be listed more than once
                            {nodeID      UserID, -- Node ID of the node--
                             nodeUpdate
                               CHOICE {addRecord      NodeRecord,
                                       replaceRecord  NodeRecord,
                                       removeRecord   NULL,
                                       ...}},
                        ...},
              rosterInstanceNumber  INTEGER(0..65535),
              nodesAdded            BOOLEAN,
              -- Nodes have been added since last instance
              nodesRemoved          BOOLEAN,
              -- Nodes have been removed since last instance
              ...},
  applicationInformation
    SET (SIZE (0..65535)) OF
      SEQUENCE
       -- One for each Application Protocol Session;
      -- all Application Protocol Sessions if full refresh;
      -- no Application Protocol Session shall be
      -- listed more than once
      {sessionKey                   SessionKey,
       applicationRecordList
         CHOICE {noChange  NULL,
                 refresh
                   SET (SIZE (0..65535)) OF
                     SEQUENCE
                      -- One for each node with the
                     -- Application Protocol Session enrolled;
                     -- no node shall be listed more than once
                     {nodeID             UserID,
                      -- Node ID of node
                      entityID           EntityID,
                      -- ID for this Application Protocol Entity at this node
                      applicationRecord  ApplicationRecord},
                 update
                   SET (SIZE (1..65536)) OF
                     SEQUENCE
                      -- One for each node modifying its Application Record;
                     -- no node shall be listed more than once
                     {nodeID             UserID,
                      -- Node ID of node
                      entityID           EntityID,
                      -- ID for this Application Protocol Entity at this node
                      applicationUpdate
                        CHOICE {addRecord      ApplicationRecord,
                                replaceRecord  ApplicationRecord,
                                removeRecord   NULL,
                                ...}},
                 ...},
       applicationCapabilitiesList
         CHOICE {noChange  NULL,
                 refresh
                   SET OF
                     SEQUENCE {capabilityID      CapabilityID,
                               capabilityClass   CapabilityClass,
                               numberOfEntities  INTEGER(1..65536),
                               -- Number of Application Protocol Entities
                               -- which issued the capability
                               ...},
                 ...},
       rosterInstanceNumber         INTEGER(0..65535),
       peerEntitiesAdded            BOOLEAN,
       -- Peer Entities have been added since last instance
       peerEntitiesRemoved          BOOLEAN,
       -- Peer Entities have been removed since last instance
       ...},
  ...
}

ApplicationInvokeIndication ::=
  SEQUENCE { -- MCS-Send-Data or MCS-Uniform-Send-Data
  -- on GCC-Broadcast-Channel or Node ID Channel
  applicationProtocolEntiyList
    SET (SIZE (1..65536)) OF ApplicationInvokeSpecifier,
  destinationNodes              SET (SIZE (1..65536)) OF UserID OPTIONAL,
  -- List of Node IDs,
  -- not present if destined for all nodes
  ...
}

RegistryRegisterChannelRequest ::=
  SEQUENCE { -- MCS-Send-Data on Node ID Channel of Top GCC
  entityID   EntityID,
  key        RegistryKey,
  channelID  DynamicChannelID,
  ...
}

RegistryAssignTokenRequest ::=
  SEQUENCE { -- MCS-Send-Data on Node ID Channel of Top GCC
  entityID  EntityID,
  key       RegistryKey,
  ...
}

RegistrySetParameterRequest ::=
  SEQUENCE { -- MCS-Send-Data on Node ID Channel of Top GCC
  entityID            EntityID,
  key                 RegistryKey,
  parameter           OCTET STRING(SIZE (0..64)),
  modificationRights  RegistryModificationRights OPTIONAL,
  ...
}

RegistryRetrieveEntryRequest ::=
  SEQUENCE { -- MCS-Send-Data on Node ID Channel of Top GCC
  entityID  EntityID,
  key       RegistryKey,
  ...
}

RegistryDeleteEntryRequest ::=
  SEQUENCE { -- MCS-Send-Data on Node ID Channel of Top GCC
  entityID  EntityID,
  key       RegistryKey,
  ...
}

RegistryMonitorEntryRequest ::=
  SEQUENCE { -- MCS-Send-Data on Node ID Channel of Top GCC
  entityID  EntityID,
  key       RegistryKey,
  ...
}

RegistryMonitorEntryIndication ::=
  SEQUENCE { -- MCS-Uniform-Send-Data on GCC-Broadcast-Channel
  key                 RegistryKey,
  item                RegistryItem,
  -- Contents: channel, token, parameter, or empty
  owner               RegistryEntryOwner,
  modificationRights  RegistryModificationRights OPTIONAL,
  ...
}

RegistryAllocateHandleRequest ::=
  SEQUENCE { -- MCS-Send-Data on Node ID Channel of Top GCC
  entityID         EntityID,
  numberOfHandles  INTEGER(1..1024),
  ...
}

RegistryAllocateHandleResponse ::=
  SEQUENCE { -- MCS-Send-Data on Node ID Channel of requester
  entityID         EntityID,
  numberOfHandles  INTEGER(1..1024),
  firstHandle      Handle,
  result           ENUMERATED {successful(0), noHandlesAvailable(1), ...
                               },
  ...
}

RegistryResponse ::=
  SEQUENCE { -- MCS-Send-Data on Node ID Channel of requester
  entityID            EntityID,
  -- Entity ID of the requesting Application Protocol Entity
  primitiveType
    ENUMERATED {registerChannel(0), assignToken(1), setParameter(2),
                retrieveEntry(3), deleteEntry(4), monitorEntry(5), ...
                },
  key                 RegistryKey,
  -- Database index
  item                RegistryItem,
  -- Contents: channel, token, parameter, or vacant
  owner               RegistryEntryOwner,
  modificationRights  RegistryModificationRights OPTIONAL,
  result
    ENUMERATED {successful(0), belongsToOther(1), tooManyEntries(2),
                inconsistentType(3), entryNotFound(4), entryAlreadyExists(5),
                invalidRequester(6), ...
                },
  ...
}

ConductorAssignIndication ::=
  SEQUENCE { -- MCS-Uniform-Send-Data on GCC-Broadcast-Channel
  conductingNode  UserID,
  ...
}

ConductorReleaseIndication ::=
  SEQUENCE { -- MCS-Uniform-Send-Data on GCC-Broadcast-Channel
  -- No parameters
  ...
}

ConductorPermissionAskIndication ::=
  SEQUENCE { -- MCS-Uniform-Send-Data on GCC-Broadcast-Channel
  grantFlag  BOOLEAN,
  -- TRUE to request permission grant, FALSE to release
  ...
}

ConductorPermissionGrantIndication ::=
  SEQUENCE { -- MCS-Uniform-Send-Data on GCC-Broadcast-Channel
  permissionList  SEQUENCE (SIZE (0..65535)) OF UserID,
  -- Node ID of nodes granted permission
  waitingList     SEQUENCE (SIZE (1..65536)) OF UserID OPTIONAL,
  -- Node ID of nodes waiting form permission
  ...
}

ConferenceTimeRemainingIndication ::=
  SEQUENCE { -- MCS-Send-Data on GCC-Broadcast-Channel
  timeRemaining  Time,
  nodeID         UserID OPTIONAL,
  ...
}

ConferenceTimeInquireIndication ::=
  SEQUENCE { -- MCS-Send-Data on GCC-Convener-Channel
  nodeSpecificTimeFlag  BOOLEAN,
  -- FALSE for conference-wide, TRUE for node-specific
  ...
}

ConferenceTimeExtendIndication ::=
  SEQUENCE { -- MCS-Send-Data on GCC-Convener-Channel
  timeToExtend          Time,
  nodeSpecificTimeFlag  BOOLEAN,
  -- FALSE for conference-wide, TRUE for node-specific
  ...
}

ConferenceAssistanceIndication ::=
  SEQUENCE { -- MCS-Uniform-Send-Data on GCC-Broadcast-Channel
  userData  UserData OPTIONAL,
  ...
}

TextMessageIndication ::= SEQUENCE { -- MCS-Send-Data or MCS-Uniform-Send-Data
  message  TextString,
  -- on GCC-Broadcast-Channel or Node ID Channel
  ...
}

RosterRefreshRequest ::= SEQUENCE {
  nodeID                UserID,
  nodeCategory          NodeCategory,
  fullRefresh           BOOLEAN,
  sendConferenceRoster  BOOLEAN OPTIONAL,
  applicationList
    SEQUENCE {applicationKeyList
                SET OF
                  SEQUENCE {applicationProtocolKey  Key,
                            nonStandardParameter
                              NonStandardParameter OPTIONAL,
                            ...},
              nonStandardParameter  NonStandardParameter OPTIONAL,
              ...} OPTIONAL,
  sessionList
    SEQUENCE {sessionKeyList
                SET OF
                  SEQUENCE {sessionKey            SessionKey,
                            nonStandardParameter  NonStandardParameter OPTIONAL,
                            ...},
              nonStandardParameter  NonStandardParameter OPTIONAL,
              ...} OPTIONAL,
  nonStandardParameter  NonStandardParameter OPTIONAL,
  ...
}

FunctionNotSupportedResponse ::= SEQUENCE {request  RequestPDU
}

NonStandardPDU ::= SEQUENCE {data  NonStandardParameter,
                             ...
}

-- ==========================================================================
-- Part 3:  Messages sent as MCS-Connect-Provider user data
-- ==========================================================================
ConnectData ::= SEQUENCE {
  t124Identifier  Key,
  -- This shall be set to the value {itu-t recommendation t 124 version(0) 1}
  connectPDU      OCTET STRING
}

ConnectGCCPDU ::= CHOICE {
  conferenceCreateRequest   ConferenceCreateRequest,
  conferenceCreateResponse  ConferenceCreateResponse,
  conferenceQueryRequest    ConferenceQueryRequest,
  conferenceQueryResponse   ConferenceQueryResponse,
  conferenceJoinRequest     ConferenceJoinRequest,
  conferenceJoinResponse    ConferenceJoinResponse,
  conferenceInviteRequest   ConferenceInviteRequest,
  conferenceInviteResponse  ConferenceInviteResponse,
  ...
}

-- ============================================================================
-- Part 4:  Messages sent using MCS-Send-Data or MCS-Uniform-Send-Data
-- ============================================================================
GCCPDU ::= CHOICE {
  request     RequestPDU,
  response    ResponsePDU,
  indication  IndicationPDU
}

RequestPDU ::= CHOICE {
  conferenceJoinRequest           ConferenceJoinRequest,
  conferenceAddRequest            ConferenceAddRequest,
  conferenceLockRequest           ConferenceLockRequest,
  conferenceUnlockRequest         ConferenceUnlockRequest,
  conferenceTerminateRequest      ConferenceTerminateRequest,
  conferenceEjectUserRequest      ConferenceEjectUserRequest,
  conferenceTransferRequest       ConferenceTransferRequest,
  registryRegisterChannelRequest  RegistryRegisterChannelRequest,
  registryAssignTokenRequest      RegistryAssignTokenRequest,
  registrySetParameterRequest     RegistrySetParameterRequest,
  registryRetrieveEntryRequest    RegistryRetrieveEntryRequest,
  registryDeleteEntryRequest      RegistryDeleteEntryRequest,
  registryMonitorEntryRequest     RegistryMonitorEntryRequest,
  registryAllocateHandleRequest   RegistryAllocateHandleRequest,
  nonStandardRequest              NonStandardPDU,
  ...
}

ResponsePDU ::= CHOICE {
  conferenceJoinResponse          ConferenceJoinResponse,
  conferenceAddResponse           ConferenceAddResponse,
  conferenceLockResponse          ConferenceLockResponse,
  conferenceUnlockResponse        ConferenceUnlockResponse,
  conferenceTerminateResponse     ConferenceTerminateResponse,
  conferenceEjectUserResponse     ConferenceEjectUserResponse,
  conferenceTransferResponse      ConferenceTransferResponse,
  registryResponse                RegistryResponse,
  registryAllocateHandleResponse  RegistryAllocateHandleResponse,
  functionNotSupportedResponse    FunctionNotSupportedResponse,
  nonStandardResponse             NonStandardPDU,
  ...
}

IndicationPDU ::= CHOICE {
  userIDIndication                    UserIDIndication,
  conferenceLockIndication            ConferenceLockIndication,
  conferenceUnlockIndication          ConferenceUnlockIndication,
  conferenceTerminateIndication       ConferenceTerminateIndication,
  conferenceEjectUserIndication       ConferenceEjectUserIndication,
  conferenceTransferIndication        ConferenceTransferIndication,
  rosterUpdateIndication              RosterUpdateIndication,
  applicationInvokeIndication         ApplicationInvokeIndication,
  registryMonitorEntryIndication      RegistryMonitorEntryIndication,
  conductorAssignIndication           ConductorAssignIndication,
  conductorReleaseIndication          ConductorReleaseIndication,
  conductorPermissionAskIndication    ConductorPermissionAskIndication,
  conductorPermissionGrantIndication  ConductorPermissionGrantIndication,
  conferenceTimeRemainingIndication   ConferenceTimeRemainingIndication,
  conferenceTimeInquireIndication     ConferenceTimeInquireIndication,
  conferenceTimeExtendIndication      ConferenceTimeExtendIndication,
  conferenceAssistanceIndication      ConferenceAssistanceIndication,
  textMessageIndication               TextMessageIndication,
  nonStandardIndication               NonStandardPDU,
  ...
}

END

-- Generated by Asnp, the ASN.1 pretty-printer of France Telecom R&D

"""


import io
import enum

class LARGE_POINTER(enum.Enum):
	FLAG_96x96 = 0x00000001 #Mouse pointer shapes of up to 96x96 pixels in size are supported.
	FLAG_384x384 = 0x00000002 #Mouse pointer shapes of up to 384x384 pixels in size, and the Fast-Path Large Pointer Update (section 2.2.9.1.2.1.11), are supported.
	FLAG_3_UNK = 0x00000003

class TS_LARGE_POINTER_CAPABILITYSET:
	def __init__(self):
		self.largePointerSupportFlags:LARGE_POINTER = None

	def to_bytes(self):
		t = self.largePointerSupportFlags.value.to_bytes(2, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_LARGE_POINTER_CAPABILITYSET.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_LARGE_POINTER_CAPABILITYSET()
		msg.largePointerSupportFlags = LARGE_POINTER(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		return msg

	def __repr__(self):
		t = '==== TS_LARGE_POINTER_CAPABILITYSET ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import os
from hashlib import md5, sha1
from arc4 import ARC4 as RC4

class RDPCryptoLayer:
	def __init__(self, ServerRandom, keysize = 128, is_FIPS = False):
		self.keysize = keysize
		self.ClientRandom:bytes = os.urandom(32) #os.urandom(keysize//8)
		self.ServerRandom:bytes = ServerRandom
		self.PacketCount:int = 0
		self.PacketCountServer:int = 0
		self.PreMasterSecret:bytes = self.First192Bits(self.ClientRandom) + self.First192Bits(self.ServerRandom)
		self.MasterSecret:bytes = None
		self.SessionKeyBlob:bytes = None
		self.MACKey128:bytes = None
		self.InitialServerEncryptKey128:bytes = None
		self.InitialServerDecryptKey128:bytes = None
		self.InitialClientDecryptKey128:bytes = None
		self.InitialClientEncryptKey128:bytes = None
		self.CurrentClientDecryptKey128:bytes = None
		self.CurrentClientEncryptKey128:bytes = None

		self.MACKey40:bytes = None
		self.InitialServerEncryptKey40:bytes = None
		self.InitialServerDecryptKey40:bytes = None
		self.InitialClientEncryptKey40:bytes = None
		self.InitialClientDecryptKey40:bytes = None
		self.CurrentClientDecryptKey40:bytes = None
		self.CurrentClientEncryptKey40:bytes = None

		self.MACKey56:bytes = None
		self.InitialServerEncryptKey56:bytes = None
		self.InitialServerDecryptKey56:bytes = None
		self.InitialClientEncryptKey56:bytes = None
		self.InitialClientDecryptKey56:bytes = None
		self.CurrentClientDecryptKey56:bytes = None
		self.CurrentClientEncryptKey56:bytes = None



		self.client_crypto_enc = None
		self.client_crypto_dec = None
		self.server_crypto_enc = None
		self.server_crypto_dec = None
		self.current_mac = None
		self.use_encrypted_mac = False #initial

		self.setup()

	def client_enc(self, data:bytes):
		"""Use this if you are a client and wanna send data to server"""
		#print('PacketCount %s' % self.PacketCount)
		if self.PacketCount != 0 and self.PacketCount % 4096 == 0:
			#print('Updating keys!')
			self.UpdateKeys(False)
		self.PacketCount += 1
		return self.client_crypto_enc.encrypt(data)
	
	def client_dec(self, data:bytes):
		"""Use this if you are a client and just got data from the server"""
		#print('PacketCountServer %s' % self.PacketCountServer)
		if self.PacketCountServer != 0 and self.PacketCountServer % 4096 == 0:
			#print('Updating keys!')
			self.UpdateKeys(True)
		self.PacketCountServer += 1
		return self.client_crypto_dec.encrypt(data)
	
	def server_enc(self, data:bytes):
		return self.server_crypto_enc.encrypt(data)
	
	def server_dec(self, data:bytes):
		return self.server_crypto_dec.encrypt(data)
	
	def calc_mac(self, data:bytes):
		# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/7c61b54e-f6cd-4819-a59a-daf200f6bf94
		Pad1 = b'\x36'*40 # repeated 40 times to give 320 bits
		Pad2 = b'\x5C'*48 # repeated 48 times to give 384 bits

		DataLength = len(data).to_bytes(4, byteorder='little', signed=False)
		
		SHAComponent = sha1(self.current_mac + Pad1 + DataLength + data).digest()
		return self.First64Bits(md5(self.current_mac + Pad2 + SHAComponent).digest())

	def calc_salted_mac(self, data:bytes, is_server:bool = False):
		# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/f390586a-3fdd-4a96-bbe0-a31575e92056
		Pad1 = b'\x36'*40 # repeated 40 times to give 320 bits
		Pad2 = b'\x5C'*48 # repeated 48 times to give 384 bits

		DataLength = len(data).to_bytes(4, byteorder='little', signed=False)
		
		if is_server is True:
			EncryptionCount = (self.PacketCountServer-1).to_bytes(4, byteorder='little', signed=False)
		else:
			EncryptionCount = self.PacketCount.to_bytes(4, byteorder='little', signed=False)

		SHAComponent = sha1(self.current_mac + Pad1 + DataLength + data + EncryptionCount).digest()
		return self.First64Bits(md5(self.current_mac + Pad2 + SHAComponent).digest())


	def setup(self):
		self.MasterSecret = self.PreMasterHash(b'\x41') + self.PreMasterHash(b'\x42\x42') + self.PreMasterHash(b'\x43\x43\x43')
		self.SessionKeyBlob = self.MasterHash(b'\x58') + self.MasterHash(b'\x59\x59') + self.MasterHash(b'\x5A\x5A\x5A')
		self.MACKey128 = self.First128Bits(self.SessionKeyBlob)

		self.InitialServerEncryptKey128 = self.FinalHash(self.Second128Bits(self.SessionKeyBlob))
		self.InitialServerDecryptKey128 = self.FinalHash(self.Third128Bits(self.SessionKeyBlob))
		self.InitialClientDecryptKey128 = self.FinalHash(self.Second128Bits(self.SessionKeyBlob))
		self.InitialClientEncryptKey128 = self.FinalHash(self.Third128Bits(self.SessionKeyBlob))

		self.MACKey40 = b'\xD1\x26\x9E' + self.Last40Bits(self.First64Bits(self.MACKey128))
  
		self.InitialServerEncryptKey40 = b'\xD1\x26\x9E' + self.Last40Bits(self.First64Bits(self.InitialServerEncryptKey128))
		self.InitialServerDecryptKey40 = b'\xD1\x26\x9E' + self.Last40Bits(self.First64Bits(self.InitialServerDecryptKey128))
		
		self.InitialClientEncryptKey40 = b'\xD1\x26\x9E' + self.Last40Bits(self.First64Bits(self.InitialClientEncryptKey128))
		self.InitialClientDecryptKey40 = b'\xD1\x26\x9E' + self.Last40Bits(self.First64Bits(self.InitialClientDecryptKey128))

		self.MACKey56 = b'\xD1' + self.Last56Bits(self.First64Bits(self.MACKey128))
  
		self.InitialServerEncryptKey56 = b'\xD1' + self.Last56Bits(self.First64Bits(self.InitialServerEncryptKey128))
		self.InitialServerDecryptKey56 = b'\xD1' + self.Last56Bits(self.First64Bits(self.InitialServerDecryptKey128))
		
		self.InitialClientEncryptKey56 = b'\xD1' + self.Last56Bits(self.First64Bits(self.InitialClientEncryptKey128))
		self.InitialClientDecryptKey56 = b'\xD1' + self.Last56Bits(self.First64Bits(self.InitialClientDecryptKey128))

		if self.keysize == 128:
			self.current_mac = self.MACKey128
			self.server_crypto_enc = RC4(self.InitialServerEncryptKey128)
			self.server_crypto_dec = RC4(self.InitialServerDecryptKey128)
			self.client_crypto_dec = RC4(self.InitialClientDecryptKey128)
			self.client_crypto_enc = RC4(self.InitialClientEncryptKey128)

			self.CurrentClientDecryptKey128 = self.InitialClientDecryptKey128
			self.CurrentClientEncryptKey128 = self.InitialClientEncryptKey128
		
		if self.keysize == 56:
			self.current_mac = self.MACKey56
			self.server_crypto_enc = RC4(self.InitialServerEncryptKey56)
			self.server_crypto_dec = RC4(self.InitialServerDecryptKey56)
			self.client_crypto_dec = RC4(self.InitialClientDecryptKey56)
			self.client_crypto_enc = RC4(self.InitialClientEncryptKey56)

			self.CurrentClientDecryptKey56 = self.InitialClientDecryptKey56
			self.CurrentClientEncryptKey56 = self.InitialClientEncryptKey56
			
		
		if self.keysize == 40:
			self.current_mac = self.MACKey40
			self.server_crypto_enc = RC4(self.InitialServerEncryptKey40)
			self.server_crypto_dec = RC4(self.InitialServerDecryptKey40)
			self.client_crypto_dec = RC4(self.InitialClientDecryptKey40)
			self.client_crypto_enc = RC4(self.InitialClientEncryptKey40)

			self.CurrentClientDecryptKey40 = self.InitialClientDecryptKey40
			self.CurrentClientEncryptKey40 = self.InitialClientEncryptKey40

			

	def UpdateKeys(self, is_server:bool = False):
		# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/a4551a30-db5c-437a-a170-a3934cf8770e
		# The encryption and the decryption keys are updated after 4,096 packets have been sent or received.

		Pad1 = b'\x36' * 40 #repeated 40 times to give 320 bits
		Pad2 = b'\x5C' * 48 #repeated 48 times to give 384 bits

		if self.keysize == 128:
			if is_server is True:
				InitialEncryptKey = self.InitialClientDecryptKey128
				CurrentEncryptKey = self.CurrentClientDecryptKey128
			else:
				InitialEncryptKey = self.InitialClientEncryptKey128
				CurrentEncryptKey = self.CurrentClientEncryptKey128

		if self.keysize == 56:
			if is_server is True:
				InitialEncryptKey = self.First64Bits(self.InitialClientDecryptKey56)
				CurrentEncryptKey = self.First64Bits(self.CurrentClientDecryptKey56)
			else:
				InitialEncryptKey = self.First64Bits(self.InitialClientEncryptKey56)
				CurrentEncryptKey = self.First64Bits(self.CurrentClientEncryptKey56)
		
		if self.keysize == 40:
			if is_server is True:
				InitialEncryptKey = self.First64Bits(self.InitialClientDecryptKey40)
				CurrentEncryptKey = self.First64Bits(self.CurrentClientDecryptKey40)
			else:
				InitialEncryptKey = self.First64Bits(self.InitialClientEncryptKey40)
				CurrentEncryptKey = self.First64Bits(self.CurrentClientEncryptKey40)
		
		SHAComponent = sha1(InitialEncryptKey + Pad1 + CurrentEncryptKey).digest()
		TempKey128 = md5(InitialEncryptKey + Pad2 + SHAComponent).digest()
		TempKey64 = self.First64Bits(TempKey128)
		
		if self.keysize == 128:
			ctx = RC4(TempKey128)
			if is_server is True:
				self.CurrentClientDecryptKey128 = ctx.encrypt(TempKey128)
				self.client_crypto_dec = RC4(self.CurrentClientDecryptKey128)
			else:
				self.CurrentClientEncryptKey128 = ctx.encrypt(TempKey128)
				self.client_crypto_enc = RC4(self.CurrentClientEncryptKey128)
		else:
			ctx = RC4(TempKey64)
			PreSaltKey = ctx.encrypt(TempKey64)
			NewEncryptKey40 = 0xD1269E + self.Last40Bits(PreSaltKey)
			NewEncryptKey56 = 0xD1 + self.Last56Bits(PreSaltKey)

			if self.keysize == 56:
				if is_server is True:
					self.CurrentClientDecryptKey56 = ctx.encrypt(NewEncryptKey56)
					self.client_crypto_dec = RC4(self.CurrentClientDecryptKey56)
				else:
					self.CurrentClientEncryptKey56 = ctx.encrypt(NewEncryptKey56)
					self.client_crypto_enc = RC4(self.CurrentClientEncryptKey56)
			
			if self.keysize == 40:
				if is_server is True:
					self.CurrentClientDecryptKey40 = ctx.encrypt(NewEncryptKey40)
					self.client_crypto_dec = RC4(self.CurrentClientDecryptKey40)
				else:
					self.CurrentClientEncryptKey40 = ctx.encrypt(NewEncryptKey40)
					self.client_crypto_enc = RC4(self.CurrentClientEncryptKey40)


	def Last56Bits(self, x:bytes):
		return x[-7:]

	def First64Bits(self, x:bytes):
		return x[:8]

	def Last40Bits(self, x:bytes):
		return x[-5:]

	def Second128Bits(self, x:bytes):
		return x[16:32]
	
	def Third128Bits(self, x:bytes):
		return x[32:48]

	def First192Bits(self, x:bytes):
		return x[:24]
	
	def First128Bits(self, x:bytes):
		return x[:16]
	
	def SaltedHash(self, S:bytes, I:bytes):
		return md5(S + sha1(I + S + self.ClientRandom + self.ServerRandom).digest()).digest()
	
	def MasterHash(self, I:bytes):
		return self.SaltedHash(self.MasterSecret, I)

	def FinalHash(self, K:bytes):
		return md5(K + self.ClientRandom + self.ServerRandom).digest()
	
	def PreMasterHash(self, x:bytes):
		return self.SaltedHash(self.PreMasterSecret, x)
	

	



import enum


class VIDEO_FORMAT(enum.Enum):
	RAW = 0 # raw bytes
	PIL = 1 # Pillow (PIL) Image object
	PNG = 2 # PNG file bytes
	QT5 = 3 # pyQt5

class MOUSEBUTTON(enum.Enum):
	MOUSEBUTTON_HOVER = -1
	MOUSEBUTTON_LEFT = 0
	MOUSEBUTTON_RIGHT = 1
	MOUSEBUTTON_MIDDLE = 2
	MOUSEBUTTON_WHEEL_UP = 3
	MOUSEBUTTON_WHEEL_DOWN = 4
	MOUSEBUTTON_5 = 5
	MOUSEBUTTON_6 = 6
	MOUSEBUTTON_7 = 7
	MOUSEBUTTON_8 = 8
	MOUSEBUTTON_9 = 9
	MOUSEBUTTON_10 = 10
	MOUSEBUTTON_11 = 11
	MOUSEBUTTON_12 = 12
	MOUSEBUTTON_13 = 13
	MOUSEBUTTON_14 = 14
	MOUSEBUTTON_15 = 15
	MOUSEBUTTON_16 = 16
	MOUSEBUTTON_17 = 17
	MOUSEBUTTON_18 = 18
	MOUSEBUTTON_19 = 19
	MOUSEBUTTON_20 = 20
	MOUSEBUTTON_21 = 21
	MOUSEBUTTON_22 = 22
	MOUSEBUTTON_23 = 23
	MOUSEBUTTON_24 = 24
	MOUSEBUTTON_25 = 25
	MOUSEBUTTON_26 = 26
	MOUSEBUTTON_27 = 27
	MOUSEBUTTON_28 = 28
	MOUSEBUTTON_29 = 29
	MOUSEBUTTON_30 = 30


import io
import enum

class TS_BITMAPCACHE_CAPABILITYSET:
	def __init__(self):
		self.pad:bytes = b'\x00' * 24
		self.Cache0Entries:int = 0
		self.Cache0MaximumCellSize:int = 0
		self.Cache1Entries:int = 0
		self.Cache1MaximumCellSize:int = 0
		self.Cache2Entries:int = 0
		self.Cache2MaximumCellSize:int = 0

	def to_bytes(self):
		t  = self.pad
		t += self.Cache0Entries.to_bytes(2, byteorder='little', signed = False)
		t += self.Cache0MaximumCellSize.to_bytes(2, byteorder='little', signed = False)
		t += self.Cache1Entries.to_bytes(2, byteorder='little', signed = False)
		t += self.Cache1MaximumCellSize.to_bytes(2, byteorder='little', signed = False)
		t += self.Cache2Entries.to_bytes(2, byteorder='little', signed = False)
		t += self.Cache2MaximumCellSize.to_bytes(32, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_BITMAPCACHE_CAPABILITYSET.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_BITMAPCACHE_CAPABILITYSET()
		msg.pad = buff.read(24)
		msg.Cache0Entries = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.Cache0MaximumCellSize = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.Cache1Entries = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.Cache2Entries = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.Cache2MaximumCellSize = int.from_bytes(buff.read(2), byteorder='little', signed = False)		
		return msg

	def __repr__(self):
		t = '==== TS_BITMAPCACHE_CAPABILITYSET ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t



import asyncio

class TCPStream:
	@staticmethod
	async def from_tcpsocket(tcpsocket):
		# tcpsocket must be running at this point
		in_queue = tcpsocket.in_queue
		out_queue = tcpsocket.out_queue
		closed_event = asyncio.Event()
		if hasattr(tcpsocket, 'disconnected'):
			closed_event = tcpsocket.disconnected

			
		writer = TCPStreamWriter(out_queue, closed_event)
		reader = TCPStreamReader(in_queue, closed_event)
		await writer.run()
		await reader.run()
		
		return reader, writer


class TCPStreamWriter:
	def __init__(self, out_queue, closed_event):
		self.out_queue = out_queue
		self.closed_event = closed_event

	def write(self, data):
		self.out_queue.put_nowait(data)

	def close(self):
		self.out_queue.put_nowait(b'')
		self.closed_event.set()

	async def drain(self):
		await asyncio.sleep(0)
		return

	async def run(self):
		return

class TCPStreamReader:
	def __init__(self, in_queue, closed_event):
		self.wsnet_reader_type = None
		self.in_queue = in_queue
		self.closed_event = closed_event
		self.buffer = b''
		self.data_in_evt = []
		self.err = None

		self.__lock = asyncio.Lock()

	async def run(self):
		return

	async def read(self, n = -1):
		try:
			if self.closed_event.is_set():
				return b''

			if self.__lock.locked() is True:
				raise Exception("Another operation is already in progress")
			
			if  len(self.buffer) > 0 and len(self.buffer) >= n:
				if n == -1:
					temp = self.buffer
					self.buffer = b''
				else:
					temp = self.buffer[:n]
					self.buffer = self.buffer[n:]
			
			else:
				async with self.__lock:
					if n == -1:
						res, self.err = await self.in_queue.get()
						self.buffer = b''

					while len(self.buffer) > 0 and len(self.buffer) >= n:
						res, self.err = await self.in_queue.get()
						if self.err is not None:
							if res is not None:
								self.buffer += res
							raise self.err
						self.buffer += res
					
					if n == -1:
						temp = self.buffer
						self.buffer = b''
						return temp
					else:
						temp = self.buffer[:n]
						self.buffer = self.buffer[n:]
						return temp
				return temp

		except Exception as e:
			#print(e)
			self.closed_event.set()
			raise

	async def readexactly(self, n):
		try:
			if n < 1:
				raise Exception('Readexactly must be a positive integer!')

			if self.__lock.locked() is True:
				raise Exception("Another operation is already in progress")
			async with self.__lock:
				while len(self.buffer) < n:
					res, self.err = await self.in_queue.get()
					if self.err is not None:
						if res is not None:
							self.buffer += res
						raise self.err
					
					self.buffer += res
				
				temp = self.buffer[:n]
				self.buffer = self.buffer[n:]
				return temp
		except Exception as e:
			#print(e)
			self.closed_event.set()
			raise

	async def readuntil(self, pattern):
		try:
			if self.closed_event.is_set():
				raise Exception('Pipe broken!')
			
			if self.__lock.locked() is True:
				raise Exception("Another operation is already in progress")
			
			async with self.__lock:
				ploc = self.buffer.find(pattern)
				while ploc == -1:
					res, self.err = await self.in_queue.get()
					if self.err is not None:
						if res is not None:
							self.buffer += res
						raise self.err
					
					self.buffer += res
					ploc = self.buffer.find(pattern)
				
				end = self.buffer.find(pattern)+len(pattern)
				temp = self.buffer[:end]
				self.buffer = self.buffer[end:]
				#print('readuntil ret %s' % temp)
				return temp

		except Exception as e:
			#print(e)
			self.closed_event.set()
			raise

	async def readline(self):
		return await self.readuntil(b'\n')


import io
import enum

class KBDFLAGS(enum.IntFlag):
	EXTENDED = 0x0100 #Indicates that the keystroke message contains an extended scancode. For enhanced 101-key and 102-key keyboards, extended keys include the right ALT and right CTRL keys on the main section of the keyboard; the INS, DEL, HOME, END, PAGE UP, PAGE DOWN and ARROW keys in the clusters to the left of the numeric keypad; and the Divide ("/") and ENTER keys in the numeric keypad.
	EXTENDED1 = 0x0200 #Used to send keyboard events triggered by the PAUSE key.A PAUSE key press and release MUST be sent as the following sequence of keyboard events:
		#CTRL (0x1D) DOWN
		#NUMLOCK (0x45) DOWN
		#CTRL (0x1D) UP
		#NUMLOCK (0x45) UP
		#The CTRL DOWN and CTRL UP events MUST both include the KBDFLAGS_EXTENDED1 flag.
	DOWN = 0x4000 #Indicates that the key was down prior to this event.
	RELEASE = 0x8000 #The absence of this flag indicates a key-down event, while its presence indicates a key-release event.

class TS_KEYBOARD_EVENT:
	def __init__(self):
		self.keyboardFlags:KBDFLAGS = None
		self.keyCode:int = None #The scancode of the key which triggered the event.
		self.pad2Octets:bytes = b'\x00\x00'

	def to_bytes(self):
		t = self.keyboardFlags.to_bytes(2, byteorder='little', signed=False)
		t += self.keyCode.to_bytes(2, byteorder='little', signed=False)
		t += self.pad2Octets
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_KEYBOARD_EVENT.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_KEYBOARD_EVENT()
		msg.keyboardFlags = KBDFLAGS(int.from_bytes(buff.read(2), byteorder='little', signed=False))
		msg.keyCode = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.pad2Octets = buff.read(2)
		return msg

	def __repr__(self):
		t = '==== TS_KEYBOARD_EVENT ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t



from aardwolf.protocol.x224.constants import TPDU_TYPE, NEG_FLAGS, SUPP_PROTOCOLS
from aardwolf.protocol.x224.client.connectionrequest import ConnectionRequest
from aardwolf.protocol.x224.server.connectionconfirm import ConnectionConfirm
from aardwolf.protocol.x224.data import Data

class X224Packet:
	@staticmethod
	def from_bytes(data: bytes):
		dtype = TPDU_TYPE(data[1] >> 4)
		return ptype2obj[dtype].from_bytes(data)

ptype2obj = {
	TPDU_TYPE.DATA : Data,
	TPDU_TYPE.CONNECTION_REQUEST : ConnectionRequest,
	TPDU_TYPE.CONNECTION_CONFIRM : ConnectionConfirm,
}



import io
import enum
from typing import List

class TS_CD_HEADER:
	def __init__(self):
		self.cbCompFirstRowSize:int = None
		self.cbCompMainBodySize:int = None
		self.cbScanWidth:int = None
		self.cbUncompressedSize:int = None

	def to_bytes(self):
		t = self.cbCompFirstRowSize.to_bytes(2, byteorder='little', signed=False)
		t += self.cbCompMainBodySize.to_bytes(2, byteorder='little', signed=False)
		t += self.cbScanWidth.to_bytes(2, byteorder='little', signed=False)
		t += self.cbUncompressedSize.to_bytes(2, byteorder='little', signed=False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_CD_HEADER.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_CD_HEADER()
		msg.cbCompFirstRowSize = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.cbCompMainBodySize = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.cbScanWidth = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.cbUncompressedSize = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		return msg

	def __repr__(self):
		t = '==== TS_CD_HEADER ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

class TS_BITMAP_FLAG(enum.IntFlag):
	BITMAP_COMPRESSION = 0x0001 #Indicates that the bitmap data is compressed. The bitmapComprHdr field MUST be present if the NO_BITMAP_COMPRESSION_HDR (0x0400) flag is not set.
	NO_BITMAP_COMPRESSION_HDR = 0x0400 #Indicates that the bitmapComprHdr field is not present (removed for bandwidth efficiency to save 8 bytes).

class TS_BITMAP_DATA:
	def __init__(self):
		self.destLeft:int = None
		self.destTop:int = None
		self.destRight:int = None
		self.destBottom:int = None
		self.width:int = None
		self.height:int = None
		self.bitsPerPixel:int = None
		self.flagsInt:int = None
		self.flags:TS_BITMAP_FLAG = None
		self.bitmapLength:int = None
		self.bitmapComprHdr:TS_CD_HEADER = None
		self.bitmapDataStream:bytes = None


	def to_bytes(self):
		data = b''
		if TS_BITMAP_FLAG.BITMAP_COMPRESSION in self.flags and TS_BITMAP_FLAG.NO_BITMAP_COMPRESSION_HDR not in self.flags:
			data += self.bitmapComprHdr.to_bytes()
		data += self.bitmapDataStream

		t = self.destLeft.to_bytes(2, byteorder='little', signed=False)
		t += self.destTop.to_bytes(2, byteorder='little', signed=False)
		t += self.destRight.to_bytes(2, byteorder='little', signed=False)
		t += self.destBottom.to_bytes(2, byteorder='little', signed=False)
		t += self.width.to_bytes(2, byteorder='little', signed=False)
		t += self.height.to_bytes(2, byteorder='little', signed=False)
		t += self.bitsPerPixel.to_bytes(2, byteorder='little', signed=False)
		t += self.flags.to_bytes(2, byteorder='little', signed=False)
		t += len(data).to_bytes(2, byteorder='little', signed=False)
		t += data
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_BITMAP_DATA.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_BITMAP_DATA()
		msg.destLeft = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.destTop = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.destRight = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.destBottom = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.width = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.height = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.bitsPerPixel = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.flagsInt = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.flags = TS_BITMAP_FLAG(msg.flagsInt)
		msg.bitmapLength = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		rl = msg.bitmapLength
		if TS_BITMAP_FLAG.NO_BITMAP_COMPRESSION_HDR not in msg.flags and TS_BITMAP_FLAG.BITMAP_COMPRESSION in msg.flags:
			msg.bitmapComprHdr = TS_CD_HEADER.from_buffer(buff)
			#rl -= 8
		msg.bitmapDataStream = buff.read(rl)
		return msg

	def __repr__(self):
		t = '==== TS_BITMAP_DATA ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

class TS_UPDATE_BITMAP_DATA:
	def __init__(self):
		self.updateType:int = 0x0001
		self.numberRectangles:int = None
		self.rectangles:List[TS_BITMAP_DATA] = []

	def to_bytes(self):
		t = self.updateType.to_bytes(2, byteorder='little', signed=False)
		t += len(self.rectangles).to_bytes(2, byteorder='little', signed=False)
		for pal in self.rectangles:
			t += pal.to_bytes()
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_UPDATE_BITMAP_DATA.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_UPDATE_BITMAP_DATA()
		msg.updateType = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.numberRectangles = int.from_bytes(buff.read(2), byteorder='little', signed=False)		
		for _ in range(msg.numberRectangles):
			msg.rectangles.append(TS_BITMAP_DATA.from_buffer(buff))
		return msg

	def __repr__(self):
		t = '==== TS_UPDATE_PALETTE_DATA ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t




from setuptools import setup, find_packages
from setuptools_rust import Binding, RustExtension
import re

VERSIONFILE="aardwolf/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
	verstr = mo.group(1)
else:
	raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))
print(find_packages())
setup(
	# Application name:
	name="aardwolf",

	# Version number (initial):
	version=verstr,

	# Application author details:
	author="Tamas Jos",
	author_email="info@skelsecprojects.com",

	# Packages
	packages=find_packages(),

	# Include additional files into the package
	include_package_data=True,


	# Details
	url="https://github.com/skelsec/aardwolf",

	zip_safe = False,
	#
	# license="LICENSE.txt",
	description="Asynchronous RDP protocol implementation",

	# long_description=open("README.txt").read(),
	python_requires='>=3.7',
	
	rust_extensions=[RustExtension("librlers", path= "aardwolf/utils/rlers/Cargo.toml", binding=Binding.PyO3)],


	install_requires=[
		'unicrypto>=0.0.9',
		'asyauth>=0.0.5',
		'asysocks>=0.2.2',
		'minikerberos>=0.3.3',
		'tqdm',
		'colorama',
		'asn1crypto',
		'asn1tools',
		'pyperclip>=1.8.2',
		'arc4>=0.3.0', #faster than cryptodome
		'Pillow>=9.0.0',		
	],
	
	classifiers=[
		"Programming Language :: Python :: 3.8",
		"Operating System :: OS Independent",
	],
	entry_points={
		'console_scripts': [
			'ardpscan = aardwolf.examples.scanners.__main__:main',
		],
	}
)



import io
from aardwolf.protocol.x224.constants import TPDU_TYPE, RESP_FLAGS, SUPP_PROTOCOLS, FAIL_CODE


class RDP_NEG_FAILURE:
	def __init__(self):
		self.type:int = 3 #RDP_NEG_FAILURE
		self.flags:int = 0
		self.length:int = 8
		self.failureCode:FAIL_CODE = None
		
	def to_bytes(self):
		t  = self.type.to_bytes(1, byteorder='little', signed = False)
		t += self.flags.to_bytes(1, byteorder='little', signed = False)
		t += self.length.to_bytes(2, byteorder='little', signed = False)
		t += self.failureCode.value.to_bytes(4, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return RDP_NEG_FAILURE.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = RDP_NEG_FAILURE()
		msg.type = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		msg.flags = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		msg.length = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.failureCode = FAIL_CODE(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		return msg

	def __repr__(self):
		t = '==== RDP_NEG_FAILURE ====\r\n'
		t += 'type :%s\r\n' % self.type
		t += 'flags :%s\r\n' % self.flags
		t += 'length :%s\r\n' % self.length
		t += 'failureCode :%s\r\n' % self.failureCode.name
		return t

class RDP_NEG_RSP:
	def __init__(self):
		self.type:int = 2 #TYPE_RDP_NEG_RSP
		self.flags:RESP_FLAGS = 0
		self.length:int = 8
		self.selectedProtocol:SUPP_PROTOCOLS = None
		
	def to_bytes(self):
		t  = self.type.to_bytes(1, byteorder='little', signed = False)
		t += self.flags.to_bytes(1, byteorder='little', signed = False)
		t += self.length.to_bytes(2, byteorder='little', signed = False)
		t += self.selectedProtocol.to_bytes(4, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return RDP_NEG_RSP.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = RDP_NEG_RSP()
		msg.type = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		msg.flags = RESP_FLAGS(int.from_bytes(buff.read(1), byteorder='little', signed = False))
		msg.length = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.selectedProtocol = SUPP_PROTOCOLS(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		return msg

	def __repr__(self):
		t = '==== RDP_NEG_RSP ====\r\n'
		t += 'type :%s\r\n' % self.type
		t += 'flags :%s\r\n' % self.flags
		t += 'length :%s\r\n' % self.length
		t += 'selectedProtocol :%s\r\n' % self.selectedProtocol
		return t

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/18a27ef9-6f9a-4501-b000-94b1fe3c2c10
class ConnectionConfirm:
	def __init__(self):
		# fixed
		self.length = None
		self.CR = TPDU_TYPE.CONNECTION_CONFIRM
		self.CDT = 0
		self.DST_REF = 0
		self.SRC_REF = None
		self.CLASS_OPT = 0
		self.cookie = None #optional!
		# variable
		self.rdpNegData = None #optional!
		
	def to_bytes(self):
		t = (self.CR.value << 4 | self.CDT).to_bytes(1, byteorder='big', signed = False)
		t += self.DST_REF.to_bytes(2, byteorder='big', signed = False)
		t += self.SRC_REF.to_bytes(2, byteorder='big', signed = False)
		t += self.CLASS_OPT.to_bytes(1, byteorder='big', signed = False)
		if self.rdpNegData is not None:
			t += self.rdpNegData.to_bytes()
		
		t = len(t).to_bytes(1, byteorder='big', signed = False) + t
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return ConnectionConfirm.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		fixlen = int.from_bytes(buff.read(1), byteorder='big', signed = False)
		tbuff = io.BytesIO(buff.read(fixlen-1))
		msg = ConnectionConfirm()
		msg.length = fixlen
		t = int.from_bytes(tbuff.read(1), byteorder='big', signed = False)
		msg.CR = TPDU_TYPE(t >> 4)
		msg.CDT = t & 0xF
		msg.DST_REF = int.from_bytes(tbuff.read(2), byteorder='big', signed = False)
		msg.SRC_REF = int.from_bytes(tbuff.read(2), byteorder='big', signed = False)
		msg.CLASS_OPT = int.from_bytes(tbuff.read(1), byteorder='big', signed = False)
		rest = tbuff.read()
		if len(rest) > 0:
			if rest[0] == 3:
				msg.rdpNegData = RDP_NEG_FAILURE.from_bytes(rest[:8])
			elif rest[0] == 2:
				msg.rdpNegData = RDP_NEG_RSP.from_bytes(rest[:8])

		return msg

	def __repr__(self):
		t = '==== ConnectionRequest ====\r\n'
		t += 'length :%s\r\n' % self.length
		t += 'CR :%s\r\n' % self.CR
		t += 'CDT :%s\r\n' % self.CDT
		t += 'DST_REF :%s\r\n' % self.DST_REF
		t += 'DST_REF :%s\r\n' % self.DST_REF
		t += 'CLASS_OPT :%s\r\n' % self.CLASS_OPT
		t += 'rdpNegData: %s' % self.rdpNegData
		return t




import asyncio
import traceback
import enum

from aardwolf import logger
from aardwolf.channels import Channel
from aardwolf.protocol.T124.userdata.constants import ChannelOption
from aardwolf.extensions.RDPECLIP.protocol import *
from aardwolf.extensions.RDPECLIP.protocol.clipboardcapabilities import CLIPRDR_GENERAL_CAPABILITY, CB_GENERAL_FALGS
from aardwolf.protocol.channelpdu import CHANNEL_PDU_HEADER, CHANNEL_FLAG
from aardwolf.extensions.RDPECLIP.protocol.formatlist import CLIPBRD_FORMAT,CLIPRDR_SHORT_FORMAT_NAME, CLIPRDR_LONG_FORMAT_NAME
from aardwolf.commons.queuedata import RDP_CLIPBOARD_NEW_DATA_AVAILABLE, RDP_CLIPBOARD_READY, RDPDATATYPE
from aardwolf.commons.queuedata.clipboard import RDP_CLIPBOARD_DATA_TXT
from aardwolf.protocol.T128.security import TS_SECURITY_HEADER,SEC_HDR_FLAG

class CLIPBRDSTATUS(enum.Enum):
	WAITING_SERVER_INIT = enum.auto()
	CLIENT_INIT = enum.auto()
	RUNNING = enum.auto()


class RDPECLIPChannel(Channel):
	name = 'cliprdr'
	def __init__(self, iosettings):
		Channel.__init__(self, self.name, ChannelOption.INITIALIZED|ChannelOption.ENCRYPT_RDP|ChannelOption.COMPRESS_RDP|ChannelOption.SHOW_PROTOCOL)
		self.use_pyperclip = iosettings.clipboard_use_pyperclip
		self.status = CLIPBRDSTATUS.WAITING_SERVER_INIT
		self.compression_needed = False #TODO: tie it to flags
		self.supported_formats = [CLIPBRD_FORMAT.CF_UNICODETEXT] #, CLIPBRD_FORMAT.CF_HDROP
		self.server_caps = None
		self.server_general_caps = None
		self.client_general_caps_flags = CB_GENERAL_FALGS.HUGE_FILE_SUPPORT_ENABLED | CB_GENERAL_FALGS.FILECLIP_NO_FILE_PATHS | CB_GENERAL_FALGS.STREAM_FILECLIP_ENABLED #| CB_GENERAL_FALGS.USE_LONG_FORMAT_NAMES # CB_GENERAL_FALGS.CAN_LOCK_CLIPDATA | #| CB_GENERAL_FALGS.USE_LONG_FORMAT_NAMES
		self.__buffer = b''
		self.current_server_formats = {}
		self.__requested_format = None
		self.__current_clipboard_data:RDP_CLIPBOARD_DATA_TXT = None 

	async def start(self):
		try:
			if self.use_pyperclip is True:
				try:
					import pyperclip
				except ImportError:
					logger.debug('Could not import pyperclip! Copy-paste will not work!')
					self.use_pyperclip = False
				else:
					if not pyperclip.is_available():
						logger.debug("pyperclip - Copy functionality available!")
			
			return True, None
		except Exception as e:
			return None, e

	async def stop(self):
		try:				
			return True, None
		except Exception as e:
			return None, e

	async def __send_capabilities(self):
		# server sent monitor ready, now we must send our capabilites
		try:
			# sending capabilities
			gencap = CLIPRDR_GENERAL_CAPABILITY()
			gencap.generalFlags = self.client_general_caps_flags

			caps = CLIPRDR_CAPS()
			caps.capabilitySets.append(gencap)

			msg = CLIPRDR_HEADER.serialize_packet(CB_TYPE.CB_CLIP_CAPS, 0, caps)
			await self.fragment_and_send(msg)

			## if remote drive is attached this should be sent
			# sending tempdir location
			# tempdir = CLIPRDR_TEMP_DIRECTORY()
			# tempdir.wszTempDir = 'C:\\Windows\\Temp\\'
			# msg = CLIPRDR_HEADER.serialize_packet(CB_TYPE.CB_TEMP_DIRECTORY, 0, tempdir)
			# await self.fragment_and_send(msg)

			# synchronizing formatlist
			fmtl = CLIPRDR_FORMAT_LIST()
			for reqfmt in self.supported_formats:
				fe = CLIPRDR_LONG_FORMAT_NAME()
				fe.formatId = reqfmt
				fmtl.templist.append(fe)

			self.status = CLIPBRDSTATUS.CLIENT_INIT
			msg = CLIPRDR_HEADER.serialize_packet(CB_TYPE.CB_FORMAT_LIST, 0, fmtl)
			await self.fragment_and_send(msg)
		
			return True, None
		except Exception as e:
			return None, e
	
	async def __process_in(self):
		try:
			hdr = CLIPRDR_HEADER.from_bytes(self.__buffer)

			if self.status == CLIPBRDSTATUS.RUNNING:
				if hdr.msgType == CB_TYPE.CB_FORMAT_LIST:
					fmtl = CLIPRDR_FORMAT_LIST.from_bytes(self.__buffer[8:8+hdr.dataLen], longnames=CB_GENERAL_FALGS.USE_LONG_FORMAT_NAMES in self.client_general_caps_flags, encoding='ascii' if CB_FLAG.CB_ASCII_NAMES in hdr.msgFlags else 'utf-16-le')
					self.current_server_formats = {}
					for fmte in fmtl.templist:
						self.current_server_formats[fmte.formatId] = fmte
					
					# sending back an OK
					msg = CLIPRDR_HEADER.serialize_packet(CB_TYPE.CB_FORMAT_LIST_RESPONSE, CB_FLAG.CB_RESPONSE_OK, None)
					await self.fragment_and_send(msg)

					if CLIPBRD_FORMAT.CF_UNICODETEXT in self.current_server_formats.keys():
						# pyperclip is in use and server just notified us about a new text copied, so we request the text
						# automatically
						self.__requested_format = CLIPBRD_FORMAT.CF_UNICODETEXT
						dreq = CLIPRDR_FORMAT_DATA_REQUEST()
						dreq.requestedFormatId = CLIPBRD_FORMAT.CF_UNICODETEXT
						msg = CLIPRDR_HEADER.serialize_packet(CB_TYPE.CB_FORMAT_DATA_REQUEST, 0, dreq)
						await self.fragment_and_send(msg)

						# notifying client that new data is available
						msg = RDP_CLIPBOARD_NEW_DATA_AVAILABLE()
						await self.send_user_data(msg)

				
				elif hdr.msgType == CB_TYPE.CB_FORMAT_DATA_RESPONSE:
					if hdr.msgFlags != hdr.msgFlags.CB_RESPONSE_OK:
						logger.debug('Server rejected our copy request!')
					else:
						try:
							fmtdata = CLIPRDR_FORMAT_DATA_RESPONSE.from_bytes(self.__buffer[8:8+hdr.dataLen],otype=self.__requested_format)
						
							if self.use_pyperclip is True and self.__requested_format in [CLIPBRD_FORMAT.CF_TEXT, CLIPBRD_FORMAT.CF_UNICODETEXT]:
								import pyperclip
								pyperclip.copy(fmtdata.dataobj)

							msg = RDP_CLIPBOARD_DATA_TXT()
							msg.data = fmtdata.dataobj
							msg.datatype = self.__requested_format
							await self.send_user_data(msg)
						
						except Exception as e:
							raise e
						finally:
							self.__requested_format = None
				
				elif hdr.msgType == CB_TYPE.CB_FORMAT_DATA_REQUEST:

					fmtr = CLIPRDR_FORMAT_DATA_REQUEST.from_bytes(self.__buffer[8:8+hdr.dataLen])
					if fmtr.requestedFormatId == self.__current_clipboard_data.datatype:
						resp = CLIPRDR_FORMAT_DATA_RESPONSE()
						resp.dataobj = self.__current_clipboard_data.data
						resp = resp.to_bytes(self.__current_clipboard_data.datatype)

						msg = CLIPRDR_HEADER.serialize_packet(CB_TYPE.CB_FORMAT_DATA_RESPONSE, CB_FLAG.CB_RESPONSE_OK, resp)
						await self.fragment_and_send(msg)
					
					else:
						logger.debug('Server requested a formatid which we dont have. %s' % fmtr.requestedFormatId)
						
			
			elif self.status == CLIPBRDSTATUS.WAITING_SERVER_INIT:
				# we expect either CLIPRDR_CAPS or CLIPRDR_MONITOR_READY
				if hdr.msgType == CB_TYPE.CB_CLIP_CAPS:
					self.server_caps = CLIPRDR_CAPS.from_bytes(self.__buffer[8:8+hdr.dataLen])
					self.server_general_caps = self.server_caps.capabilitySets[0] #it's always the generalflags
					logger.debug(self.server_general_caps)
				elif hdr.msgType == CB_TYPE.CB_MONITOR_READY:
					_, err = await self.__send_capabilities()
					if err is not None:
						raise err
				else:
					raise Exception('Unexpected packet type %s arrived!' % hdr.msgType.name)

				#await self.out_queue.put((data, err))
			elif self.status == CLIPBRDSTATUS.CLIENT_INIT:
				# initialization started, we already sent all necessary data
				# here we expect CB_FORMAT_LIST_RESPONSE
				if hdr.msgType == CB_TYPE.CB_FORMAT_LIST_RESPONSE:
					#this doesnt hold any data
					if CB_FLAG.CB_RESPONSE_OK in hdr.msgFlags:
						# everything was okay, now we can communicate on this channel normally
						# also we have to notify the client that they can use the keyboard now
						self.status = CLIPBRDSTATUS.RUNNING
						msg = RDP_CLIPBOARD_READY()
						await self.send_user_data(msg)

					elif CB_FLAG.CB_RESPONSE_FAIL in hdr.msgFlags:
						raise Exception('Server refused clipboard initialization!')
					else:
						raise Exception('Server sent unexpected data! %s' % hdr)
				


			self.__buffer = self.__buffer[8+hdr.dataLen:]
			return True, None
		except Exception as e:
			return None, e

	async def process_channel_data(self, data):
		channeldata = CHANNEL_PDU_HEADER.from_bytes(data)
		#print('channeldata %s' % channeldata)
		self.__buffer += channeldata.data
		if CHANNEL_FLAG.CHANNEL_FLAG_LAST in channeldata.flags:
			_, err = await self.__process_in()
			if err is not None:
				raise err

	async def fragment_and_send(self, data):
		try:
			if len(data) < 16000:
				if self.compression_needed is False:
					flags = CHANNEL_FLAG.CHANNEL_FLAG_FIRST|CHANNEL_FLAG.CHANNEL_FLAG_LAST|CHANNEL_FLAG.CHANNEL_FLAG_SHOW_PROTOCOL
					packet = CHANNEL_PDU_HEADER.serialize_packet(flags, data)
				else:
					raise NotImplementedError('Compression not implemented!')
			else:
				raise NotImplementedError('Chunked send not implemented!')
			
			sec_hdr = None
			if self.connection.cryptolayer is not None:
				sec_hdr = TS_SECURITY_HEADER()
				sec_hdr.flags = SEC_HDR_FLAG.ENCRYPT
				sec_hdr.flagsHi = 0

			await self.send_channel_data(packet, sec_hdr, None, None, False)

			return True, False
		except Exception as e:
			traceback.print_exc()
			return None,e
	

	async def process_user_data(self, data):
		#print('monitor out! %s' % data)
		if data.type == RDPDATATYPE.CLIPBOARD_DATA_TXT:
			# data in, informing the server that our clipboard has changed
			if data == self.__current_clipboard_data:
				return
					
			fmtl = CLIPRDR_FORMAT_LIST()
			for fmtid in [CLIPBRD_FORMAT.CF_UNICODETEXT]: #CLIPBRD_FORMAT.CF_TEXT, CLIPBRD_FORMAT.CF_OEMTEXT
				if CB_GENERAL_FALGS.USE_LONG_FORMAT_NAMES not in self.server_general_caps.generalFlags:
					name = CLIPRDR_LONG_FORMAT_NAME()
					name.formatId = data.datatype
				else:
					name = CLIPRDR_SHORT_FORMAT_NAME()
					name.formatId = data.datatype
				fmtl.templist.append(name)
			msg = CLIPRDR_HEADER.serialize_packet(CB_TYPE.CB_FORMAT_LIST, 0, fmtl)
			await self.fragment_and_send(msg)

			self.__current_clipboard_data = data
				
		else:
			logger.debug('Unhandled data type in! %s' % data.type)


import io
import enum

from aardwolf.protocol.x224.constants import TPDU_TYPE, NEG_FLAGS, SUPP_PROTOCOLS

class RDP_NEG_CORRELATION_INFO:
	def __init__(self):
		self.type:int = 6 #TYPE_RDP_CORRELATION_INFO
		self.flags:int = 0 #no flags defined
		self.length:int = 36
		self.correlationId:bytes = None
		self.reserved:bytes = None
		
	def to_bytes(self):
		t  = self.type.to_bytes(1, byteorder='little', signed = False)
		t += self.flags.value.to_bytes(1, byteorder='little', signed = False)
		t += self.length.to_bytes(2, byteorder='little', signed = False)
		t += self.correlationId
		t += self.reserved
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return RDP_NEG_CORRELATION_INFO.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = RDP_NEG_CORRELATION_INFO()
		msg.type = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		msg.flags = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		msg.length = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.correlationId = buff.read(16)
		msg.reserved = buff.read(16)
		return msg

	def __repr__(self):
		t = '==== RDP_NEG_CORRELATION_INFO ====\r\n'
		t += 'type :%s\r\n' % self.type
		t += 'flags :%s\r\n' % self.flags
		t += 'length :%s\r\n' % self.length
		t += 'correlationId :%s\r\n' % self.correlationId
		t += 'reserved :%s\r\n' % self.reserved
		return t

class RDP_NEG_REQ:
	def __init__(self):
		self.type = 1
		self.flags:NEG_FLAGS = None
		self.length:int = 8
		self.requestedProtocols:SUPP_PROTOCOLS = None
		
	def to_bytes(self):
		t  = self.type.to_bytes(1, byteorder='little', signed = False)
		t += self.flags.to_bytes(1, byteorder='little', signed = False)
		t += self.length.to_bytes(2, byteorder='little', signed = False)
		t += self.requestedProtocols.to_bytes(4, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return RDP_NEG_REQ.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = RDP_NEG_REQ()
		msg.type = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		msg.flags = NEG_FLAGS(int.from_bytes(buff.read(1), byteorder='little', signed = False))
		msg.length = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.requestedProtocols = SUPP_PROTOCOLS(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		return msg

	def __repr__(self):
		t = '==== RDP_NEG_REQ ====\r\n'
		t += 'type :%s\r\n' % self.type
		t += 'flags :%s\r\n' % self.flags
		t += 'length :%s\r\n' % self.length
		t += 'requestedProtocols :%s\r\n' % self.requestedProtocols
		return t

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/18a27ef9-6f9a-4501-b000-94b1fe3c2c10
class ConnectionRequest:
	def __init__(self):
		# fixed
		self.length = None
		self.CR = TPDU_TYPE.CONNECTION_REQUEST
		self.CDT = 0
		self.DST_REF = 0
		self.SRC_REF = None
		self.CLASS_OPT = 0
		self.cookie = None #optional!
		# variable
		self.rdpNegReq = None #optional!
		self.rdpCorrelationInfo = None #optional!
		
	def to_bytes(self):
		t = (self.CR.value << 4 | self.CDT).to_bytes(1, byteorder='big', signed = False)
		t += self.DST_REF.to_bytes(2, byteorder='big', signed = False)
		t += self.SRC_REF.to_bytes(2, byteorder='big', signed = False)
		t += self.CLASS_OPT.to_bytes(1, byteorder='big', signed = False)
		if self.cookie is not None:
			t += self.cookie
		if self.rdpNegReq is not None:
			t += self.rdpNegReq.to_bytes()
		
		t = len(t).to_bytes(1, byteorder='big', signed = False) + t
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return ConnectionRequest.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		fixlen = int.from_bytes(buff.read(1), byteorder='big', signed = False)
		tbuff = io.BytesIO(buff.read(fixlen-1))
		msg = ConnectionRequest()
		msg.length = fixlen
		t = int.from_bytes(tbuff.read(1), byteorder='big', signed = False)
		msg.CR = TPDU_TYPE(t >> 4)
		msg.CDT = t & 0xF
		msg.DST_REF = int.from_bytes(tbuff.read(2), byteorder='big', signed = False)
		msg.SRC_REF = int.from_bytes(tbuff.read(2), byteorder='big', signed = False)
		msg.CLASS_OPT = int.from_bytes(tbuff.read(1), byteorder='big', signed = False)
		rest = tbuff.read()
		m = rest.find(b'\r\n')
		if m != -1:
			msg.cookie = rest[:m+2]
			rest = rest[m+2:]
		if len(rest) > 0:
			msg.rdpNegReq = RDP_NEG_REQ.from_bytes(rest[:8])
			if NEG_FLAGS.CORRELATION_INFO_PRESENT in msg.rdpNegReq.flags:
				msg.rdpCorrelationInfo = RDP_NEG_CORRELATION_INFO.from_bytes(rest[8:])

		return msg

	def __repr__(self):
		t = '==== ConnectionRequest ====\r\n'
		t += 'length :%s\r\n' % self.length
		t += 'CR :%s\r\n' % self.CR
		t += 'CDT :%s\r\n' % self.CDT
		t += 'DST_REF :%s\r\n' % self.DST_REF
		t += 'DST_REF :%s\r\n' % self.DST_REF
		t += 'CLASS_OPT :%s\r\n' % self.CLASS_OPT
		t += 'cookie :%s\r\n' % self.cookie
		t += 'rdpNegReq: %s' % self.rdpNegReq
		t += 'rdpCorrelationInfo: %s' % self.rdpCorrelationInfo
		return t



import io
import enum

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/17e80f50-d163-49de-a23b-fd6456aa472f
# TODO
class TS_BITMAPCODECS_CAPABILITYSET:
	def __init__(self):
		self.supportedBitmapCodecs:bytes = None

	def to_bytes(self):
		t = self.supportedBitmapCodecs
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_BITMAPCODECS_CAPABILITYSET.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_BITMAPCODECS_CAPABILITYSET()
		msg.supportedBitmapCodecs = buff.read()
		return msg

	def __repr__(self):
		t = '==== TS_BITMAPCODECS_CAPABILITYSET ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import copy
from asysocks.unicomm.common.scanner.common import *
from aardwolf.commons.factory import RDPConnectionFactory
from aardwolf.protocol.x224.constants import SUPP_PROTOCOLS
from aardwolf.network.tpkt import TPKTPacketizer
from aardwolf.network.x224 import X224Network
from asysocks.unicomm.client import UniClient
from aardwolf.protocol.x224.constants import SUPP_PROTOCOLS, NEG_FLAGS


class RDPCapabilitiesRes:
	def __init__(self, restrictedadmin,restrictedauth,rdp,ssl,hybrid,hybridex):
		self.restrictedadmin = restrictedadmin
		self.restrictedauth = restrictedauth
		self.rdp = rdp
		self.ssl = ssl
		self.hybrid = hybrid
		self.hybridex = hybridex

	def get_header(self):
		return ['restrictedadmin', 'restrictedauth', 'rdp', 'ssl', 'hybrid', 'hybridex']

	def to_line(self, separator = '\t'):
		return separator.join([
			str(self.restrictedadmin),
			str(self.restrictedauth),
			str(self.rdp),
			str(self.ssl),
			str(self.hybrid),
			str(self.hybridex),
		])

class RDPCapabilitiesScanner:
	def __init__(self, factory:RDPConnectionFactory):
		self.factory:RDPConnectionFactory = factory
		self.protoflags = [SUPP_PROTOCOLS.RDP, SUPP_PROTOCOLS.SSL, SUPP_PROTOCOLS.SSL|SUPP_PROTOCOLS.RDP, SUPP_PROTOCOLS.SSL|SUPP_PROTOCOLS.HYBRID, SUPP_PROTOCOLS.SSL|SUPP_PROTOCOLS.HYBRID_EX]


	
	async def scan_flags(self, target, negflag, protoflag):
		try:
			packetizer = TPKTPacketizer()
			client = UniClient(target, packetizer)
			connection = await client.connect()

			# X224 channel is on top of TPKT, performs the initial negotiation
			# between the server and our client (restricted admin mode, authentication methods etc)
			# are set here
			x224net = X224Network(connection)

			reply, err = await x224net.client_negotiate(negflag, protoflag, to_raise=False)
			if err is not None:
				raise err

			if reply.rdpNegData.type == 3:
				#server refused our flags
				return False, reply, None

			return True, reply, None
		except Exception as e:
			return False, None, e

	async def run(self, targetid, target, out_queue):
		try:
			rdp = None
			ssl = None
			hybrid = None
			hybridex = None
			restrictedadmin = None
			restrictedauth = None

			rdptarget = self.factory.create_connection_newtarget(target, self.factory.get_settings()).target

			for protoflag in self.protoflags:
				negflag = 0
				ok, reply, err = await self.scan_flags(rdptarget, negflag, protoflag)
				if err is not None:
					raise err
				if ok is True:
					if SUPP_PROTOCOLS.SSL == reply.rdpNegData.selectedProtocol:
						ssl = True
					if SUPP_PROTOCOLS.HYBRID in reply.rdpNegData.selectedProtocol:
						hybrid = True
					if SUPP_PROTOCOLS.HYBRID_EX in reply.rdpNegData.selectedProtocol:
						hybridex = True
					if SUPP_PROTOCOLS.RDP == reply.rdpNegData.selectedProtocol:
						rdp = True

			ok, reply, err = await self.scan_flags(rdptarget, NEG_FLAGS.RESTRICTED_ADMIN_MODE_REQUIRED, SUPP_PROTOCOLS.SSL|SUPP_PROTOCOLS.HYBRID)
			if err is not None:
				raise err
			restrictedadmin = ok
			ok, reply, err = await self.scan_flags(rdptarget, NEG_FLAGS.REDIRECTED_AUTHENTICATION_MODE_REQUIRED, SUPP_PROTOCOLS.SSL|SUPP_PROTOCOLS.HYBRID)
			if err is not None:
				raise err
			restrictedauth = ok

			await out_queue.put(ScannerData(target, RDPCapabilitiesRes(rdp, ssl, hybrid, hybridex, restrictedadmin, restrictedauth)))

		except Exception as e:
			await out_queue.put(ScannerError(target, e))



import io
import enum

class COMPDESK(enum.Enum):
	NOT_SUPPORTED = 0x0000 #Desktop composition services are not supported.
	SUPPORTED = 0x0001 #Desktop composition services are supported.

class TS_COMPDESK_CAPABILITYSET:
	def __init__(self):
		self.WndSupportLevel:COMPDESK = COMPDESK.NOT_SUPPORTED

	def to_bytes(self):
		t = self.WndSupportLevel.value.to_bytes(2, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_COMPDESK_CAPABILITYSET.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_COMPDESK_CAPABILITYSET()
		msg.WndSupportLevel = COMPDESK(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		return msg

	def __repr__(self):
		t = '==== TS_COMPDESK_CAPABILITYSET ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t



import io
from aardwolf.protocol.T124.userdata.constants import TS_UD_TYPE
from aardwolf.protocol.T124.userdata.clientcoredata import TS_UD_CS_CORE
from aardwolf.protocol.T124.userdata.clientsecuritydata import TS_UD_CS_SEC
from aardwolf.protocol.T124.userdata.clientnetworkdata import TS_UD_CS_NET
from aardwolf.protocol.T124.userdata.clientclusterdata import TS_UD_CS_CLUSTER
from aardwolf.protocol.T124.userdata.monitordefinition import TS_UD_CS_MONITOR
from aardwolf.protocol.T124.userdata.clientmessagechanneldata import TS_UD_CS_MCS_MSGCHANNEL
from aardwolf.protocol.T124.userdata.clientmultitransportchanneldata import TS_UD_CS_MULTITRANSPORT
from aardwolf.protocol.T124.userdata.clientmonitorextendeddata import TS_UD_CS_MONITOR_EX

from aardwolf.protocol.T124.userdata.servercoredata import TS_UD_SC_CORE
from aardwolf.protocol.T124.userdata.serversecuritydata import TS_UD_SC_SEC1
from aardwolf.protocol.T124.userdata.servernetworkdata import TS_UD_SC_NET
from aardwolf.protocol.T124.userdata.servermessagechannel import TS_UD_SC_MCS_MSGCHANNEL
from aardwolf.protocol.T124.userdata.servermultitransportchanneldata import TS_UD_SC_MULTITRANSPORT


class TS_UD:
	def __init__(self):
		self.userdata = {}

	@staticmethod
	def from_bytes(data):
		return TS_UD.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = TS_UD()
		start = buff.tell()
		buff.seek(0,2)
		end = buff.tell()
		buff.seek(start)
		while buff.tell() < end:
			print(buff.tell())
			ts = buff.tell()
			otype = TS_UD_TYPE(int.from_bytes(buff.read(2), byteorder='little', signed=False))
			olen = int.from_bytes(buff.read(2), byteorder='little', signed=False)
			buff.seek(ts)
			res.userdata[otype] = type2obj[otype].from_buffer(buff)
		return res
	
	def to_bytes(self):
		t = b''
		for k in self.userdata:
			t += self.userdata[k].to_bytes()
		return t

class TS_UD_PACKET:
	@staticmethod
	def from_bytes(data):
		olen = int.from_bytes(data[2:4], byteorder='little', signed=False)
		otype = TS_UD_TYPE(int.from_bytes(data[:2], byteorder='little', signed=False))
		return type2obj[otype].from_bytes(data[:olen])
	
	@staticmethod
	def from_buffer(buff):
		start = buff.tell()
		olen = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		otype = TS_UD_TYPE(int.from_bytes(buff.read(2), byteorder='little', signed=False))
		buff.seek(start)
		return type2obj[otype].from_buffer(buff)


type2obj = {
	TS_UD_TYPE.CS_CORE : TS_UD_CS_CORE,
	TS_UD_TYPE.CS_SECURITY : TS_UD_CS_SEC,
	TS_UD_TYPE.CS_NET : TS_UD_CS_NET,
	TS_UD_TYPE.CS_CLUSTER : TS_UD_CS_CLUSTER,
	TS_UD_TYPE.CS_MONITOR : TS_UD_CS_MONITOR,
	TS_UD_TYPE.CS_MCS_MSGCHANNEL : TS_UD_CS_MCS_MSGCHANNEL,
	TS_UD_TYPE.CS_MONITOR_EX : TS_UD_CS_MONITOR_EX,
	TS_UD_TYPE.CS_MULTITRANSPORT : TS_UD_CS_MULTITRANSPORT,
}

class TS_SC:
	def __init__(self):
		self.serverdata = {}

	@staticmethod
	def from_bytes(data):
		return TS_SC.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = TS_SC()
		start = buff.tell()
		buff.seek(0,2)
		end = buff.tell()
		buff.seek(start)
		while buff.tell() < end:
			ts = buff.tell()
			otype = TS_UD_TYPE(int.from_bytes(buff.read(2), byteorder='little', signed=False))
			olen = int.from_bytes(buff.read(2), byteorder='little', signed=False)
			buff.seek(ts)
			res.serverdata[otype] = srvtype2obj[otype].from_buffer(buff)
		return res
	
	def to_bytes(self):
		t = b''
		for k in self.serverdata:
			t += self.serverdata[k].to_bytes()
		return t

class TS_SC_PACKET:
	@staticmethod
	def from_bytes(data):
		olen = int.from_bytes(data[2:4], byteorder='little', signed=False)
		otype = TS_UD_TYPE(int.from_bytes(data[:2], byteorder='little', signed=False))
		return srvtype2obj[otype].from_bytes(data[:olen])
	
	@staticmethod
	def from_buffer(buff):
		start = buff.tell()
		olen = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		otype = TS_UD_TYPE(int.from_bytes(buff.read(2), byteorder='little', signed=False))
		buff.seek(start)
		return type2obj[otype].from_buffer(buff)


srvtype2obj = {
	TS_UD_TYPE.SC_CORE : TS_UD_SC_CORE,
	TS_UD_TYPE.SC_SECURITY : TS_UD_SC_SEC1,
	TS_UD_TYPE.SC_NET : TS_UD_SC_NET,
	TS_UD_TYPE.SC_MCS_MSGCHANNEL : TS_UD_SC_MCS_MSGCHANNEL,
	TS_UD_TYPE.SC_MULTITRANSPORT : TS_UD_SC_MULTITRANSPORT,
}


import io
import enum

class CacheFlag2(enum.IntFlag):
	PERSISTENT_KEYS_EXPECTED_FLAG = 0x0001 #Indicates that the client will send a Persistent Key List PDU during the Connection Finalization phase of the RDP Connection Sequence (see section 1.3.1.1 for an overview of the RDP Connection Sequence phases).
	ALLOW_CACHE_WAITING_LIST_FLAG = 0x0002 #Indicates that the client supports a cache waiting list. If a waiting list is supported, new bitmaps are cached on the second hit rather than the first (that is, a bitmap is sent twice before it is cached).

class TS_BITMAPCACHE_CELL_CACHE_INFO:
	def __init__(self):
		self.NumEntries:int = None
		self.persistent:bool = None

	def to_bytes(self):
		temp = self.NumEntries << 1
		temp += int(self.persistent)
		t  = temp.to_bytes(4, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_BITMAPCACHE_CELL_CACHE_INFO.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_BITMAPCACHE_CELL_CACHE_INFO()
		t = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.NumEntries = t >> 1
		msg.persistent = bool(t & 1)
		return msg

	def __repr__(self):
		t = '==== TS_BITMAPCACHE_CELL_CACHE_INFO ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

class TS_BITMAPCACHE_CAPABILITYSET_REV2:
	def __init__(self):
		self.CacheFlags:CacheFlag2 = None
		self.Pad2:bytes = b'\x00' * 1
		self.NumCellCaches:int = None
		self.BitmapCache0CellInfo:TS_BITMAPCACHE_CELL_CACHE_INFO = None
		self.BitmapCache1CellInfo:TS_BITMAPCACHE_CELL_CACHE_INFO = None
		self.BitmapCache2CellInfo:TS_BITMAPCACHE_CELL_CACHE_INFO = None
		self.BitmapCache3CellInfo:TS_BITMAPCACHE_CELL_CACHE_INFO = None
		self.BitmapCache4CellInfo:TS_BITMAPCACHE_CELL_CACHE_INFO = None
		self.Pad3:bytes = b'\x00' * 12

	def to_bytes(self):
		t  = self.CacheFlags.to_bytes(2, byteorder='little', signed = False)
		t += self.Pad2
		t += self.NumCellCaches.to_bytes(1, byteorder='little', signed = False)
		t += self.BitmapCache0CellInfo.to_bytes()
		t += self.BitmapCache1CellInfo.to_bytes()
		t += self.BitmapCache2CellInfo.to_bytes()
		t += self.BitmapCache3CellInfo.to_bytes()
		t += self.BitmapCache4CellInfo.to_bytes()
		t += self.Pad3
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_BITMAPCACHE_CAPABILITYSET_REV2.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_BITMAPCACHE_CAPABILITYSET_REV2()
		msg.CacheFlags = CacheFlag2(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.Pad2 = buff.read(1)
		msg.NumCellCaches = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		msg.BitmapCache0CellInfo = TS_BITMAPCACHE_CELL_CACHE_INFO.from_buffer(buff)
		msg.BitmapCache1CellInfo = TS_BITMAPCACHE_CELL_CACHE_INFO.from_buffer(buff)
		msg.BitmapCache2CellInfo = TS_BITMAPCACHE_CELL_CACHE_INFO.from_buffer(buff)
		msg.BitmapCache3CellInfo = TS_BITMAPCACHE_CELL_CACHE_INFO.from_buffer(buff)
		msg.BitmapCache4CellInfo = TS_BITMAPCACHE_CELL_CACHE_INFO.from_buffer(buff)
		
		msg.Pad3 = buff.read(12)
		return msg

	def __repr__(self):
		t = '==== TS_BITMAPCACHE_CAPABILITYSET_REV2 ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t